from ._optproblems import OneMax
from ._optproblems import Sphere
from ._optproblems import Schwefe1_2
from ._optproblems import Rosenbrock
from ._optproblems import Rastrigin
from ._optproblems import HighConditionedElliptic
from ._optproblems import Griewank
from ._optproblems import Ackley
from ._optproblems import Weierstrass


import numpy as np
from ._geneticprogramming import GeneticProgramming
from ..tools.transformations import scale_data
from ..tools.transformations import rank_data
from functools import partial
from ._base import TheFittest
from ._base import LastBest
from ..tools.transformations import numpy_group_by
from ..tools.generators import half_and_half


class StatisticsSelfCGP:
    def __init__(self, mode='quick'):
        self.mode = mode
        self.fittest = np.array([])
        self.fitness = np.array([])
        self.s_proba = dict()
        self.c_proba = dict()
        self.m_proba = dict()

    def update(self,
               fittest_i,
               fitness_i,
               s_proba_i, c_proba_i, m_proba_i):
        if self.mode == 'quick':
            self.fitness = np.append(self.fitness, np.max(fitness_i))
            for proba_i, archive_i in zip((s_proba_i, c_proba_i, m_proba_i),
                                          (self.s_proba, self.c_proba, self.m_proba)):
                if not len(archive_i):
                    for key, value in proba_i.items():
                        archive_i[key] = np.array(value)
                else:
                    for key, value in proba_i.items():
                        archive_i[key] = np.append(
                            archive_i[key], np.array(value))
        elif self.mode == 'full':
            self.fittest = np.append(self.fittest, fittest_i.copy())
            self.fitness = np.append(self.fitness, np.max(fitness_i))
            for proba_i, archive_i in zip((s_proba_i, c_proba_i, m_proba_i),
                                          (self.s_proba, self.c_proba, self.m_proba)):
                if not len(archive_i):
                    for key, value in proba_i.items():
                        archive_i[key] = np.array(value)
                else:
                    for key, value in proba_i.items():
                        archive_i[key] = np.append(
                            archive_i[key], np.array(value))
        else:
            raise ValueError('the "mode" must be either "quick" or "full"')
        return self


class SelfCGP(GeneticProgramming):
    '''Semenkin, Eugene & Semenkina, Maria. (2012). Self-configuring genetic programming algorithm
    with modified uniform crossover. 1-6. 10.1109/CEC.2012.6256587. '''

    def __init__(self,
                 fitness_function,
                 genotype_to_phenotype,
                 uniset,
                 iters,
                 pop_size,
                 optimal_value = None,
                 termination_error_value = 0,
                 no_increase_num = None,
                 minimization = False,
                 show_progress_each = None,
                 keep_history = False):
        super().__init__(fitness_function,
                         genotype_to_phenotype,
                         uniset, iters,
                         pop_size,
                         optimal_value,
                         termination_error_value,
                         no_increase_num,
                         minimization,
                         show_progress_each,
                         keep_history)

        self.thefittest: TheFittest
        self.stats: StatisticsSelfCGP
        self.s_sets: dict
        self.c_sets: dict
        self.m_sets: dict
        self.K: int
        self.threshold: float
        self.s_set: dict
        self.c_set: dict
        self.m_set: dict

        self.set_strategy()

    def set_strategy(self, select_opers = ['proportional',
                                                      'rank',
                                                      'tournament_3',
                                                      'tournament_5'],
                     crossover_opers = ['standart',
                                                   'one_point',
                                                   'uniform_rank7'],
                     mutation_opers = ['weak_point',
                                                  'average_point',
                                                  'strong_point',
                                                  'weak_grow',
                                                  'average_grow',
                                                  'strong_grow'],
                     tour_size_param = 2,
                     initial_population = None,
                     elitism_param = True,
                     parents_num_param = 7,
                     mutation_rate_param = 0.05,
                     K_param = 2,
                     threshold_param = 0.05,
                     max_level = 16,
                     init_level = 5):
        self.tour_size = tour_size_param
        self.initial_population = initial_population
        self.elitism = elitism_param
        self.parents_num = parents_num_param
        self.mutation_rate = mutation_rate_param
        self.K = K_param
        self.threshold = threshold_param
        self.max_level = max_level
        self.init_level = init_level

        self.update_pool()

        s_sets = {}
        for operator_name in select_opers:
            value = self.s_pool[operator_name]
            s_sets[operator_name] = value
        self.s_sets = dict(sorted(s_sets.items()))

        c_sets = {}
        for operator_name in crossover_opers:
            value = self.c_pool[operator_name]
            c_sets[operator_name] = value
        self.c_sets = dict(sorted(c_sets.items()))

        m_sets = {}
        for operator_name in mutation_opers:
            value = self.m_pool[operator_name]
            m_sets[operator_name] = value
        self.m_sets = dict(sorted(m_sets.items()))

        return self

    def create_offspring(self, population_g, fitness_scale, fitness_rank,
                         selection, crossover, mutation):
        crossover_func, quantity = self.c_sets[crossover]
        selection_func, tour_size = self.s_sets[selection]
        mutation_func, proba_down, not_scale = self.m_sets[mutation]

        indexes = selection_func(fitness_scale,
                                 fitness_rank,
                                 tour_size,
                                 quantity)

        parents = population_g[indexes]
        fitness_scale_p = fitness_scale[indexes].copy()
        fitness_rank_p = fitness_rank[indexes].copy()

        offspring_no_mutated = crossover_func(parents,
                                              fitness_scale_p,
                                              fitness_rank_p,
                                              self.max_level)

        if not_scale:
            proba = proba_down
        else:
            proba = proba_down/len(offspring_no_mutated)
        mutant = mutation_func(offspring_no_mutated,
                               self.uniset, proba, self.max_level)
        return mutant

    def choice_operators(self, proba_dict):
        operators = list(proba_dict.keys())
        proba = list(proba_dict.values())
        return np.random.choice(list(operators), self.pop_size, p=proba)

    def update_proba(self, proba, operator):
        proba[operator] += self.K/self.iters
        proba_value = np.array(list(proba.values()))
        proba_value -= self.K/(len(proba)*self.iters)
        proba_value = proba_value.clip(self.threshold, 1)
        proba_value = proba_value/proba_value.sum()
        return dict(zip(proba.keys(), proba_value))

    def find_fittest_operator(self, operators, fitness):
        keys, groups = numpy_group_by(group=fitness, by=operators)
        mean_fit = np.array(list(map(np.mean, groups)))
        return keys[np.argmax(mean_fit)]

    def fit(self):
        z_s, z_c, z_m = list(map(len, (self.s_sets, self.c_sets, self.m_sets)))
        s_proba = dict(zip(list(self.s_sets.keys()), np.full(z_s, 1/z_s)))
        m_proba = dict(zip(list(self.m_sets.keys()), np.full(z_m, 1/z_m)))
        if 'empty' in self.c_sets.keys():
            c_proba = dict(zip(list(self.c_sets.keys()),
                           np.full(z_c, 0.9/(z_c-1))))
            c_proba['empty'] = 0.1
        else:
            c_proba = dict(zip(list(self.c_sets.keys()), np.full(z_c, 1/z_c)))

        population_g = half_and_half(
            self.pop_size, self.uniset, self.init_level)
        population_ph = self.genotype_to_phenotype(population_g)
        fitness = self.evaluate(population_ph)

        fitness_scale = scale_data(fitness)
        fitness_rank = rank_data(fitness)

        self.thefittest = TheFittest().update(population_g,
                                              population_ph,
                                              fitness)
        lastbest = LastBest().update(self.thefittest.fitness)
        if self.keep_history is not None:
            self.stats = StatisticsSelfCGP(
                mode=self.keep_history).update(self.thefittest.genotype,
                                               fitness,
                                               s_proba,
                                               c_proba,
                                               m_proba)
        for i in range(self.iters-1):
            self.show_progress(i)
            if self.termitation_check(lastbest.no_increase):
                break
            else:
                s_operators = self.choice_operators(s_proba)
                c_operators = self.choice_operators(c_proba)
                m_operators = self.choice_operators(m_proba)

                partial_create_offspring = partial(self.create_offspring,
                                                   population_g,
                                                   fitness_scale,
                                                   fitness_rank)
                map_ = map(partial_create_offspring, s_operators,
                           c_operators, m_operators)
                population_g = np.array(list(map_), dtype=object)
                population_ph = self.genotype_to_phenotype(
                    population_g)
                fitness = self.evaluate(population_ph)

                if self.elitism:
                    population_g[-1], population_ph[-1], fitness[-1] = self.thefittest.get()
                fitness_scale = scale_data(fitness)
                fitness_rank = rank_data(fitness)

                s_fittest_oper = self.find_fittest_operator(
                    s_operators, fitness_scale)
                s_proba = self.update_proba(s_proba, s_fittest_oper)

                c_fittest_oper = self.find_fittest_operator(
                    c_operators, fitness_scale)
                c_proba = self.update_proba(c_proba, c_fittest_oper)

                m_fittest_oper = self.find_fittest_operator(
                    m_operators, fitness_scale)
                m_proba = self.update_proba(m_proba, m_fittest_oper)

                self.thefittest.update(population_g, population_ph, fitness)
                lastbest.update(self.thefittest.fitness)

                if self.keep_history is not None:
                    self.stats.update(self.thefittest.genotype,
                                      fitness,
                                      s_proba,
                                      c_proba,
                                      m_proba)
        return self



import numpy as np
from ..optimizers._base import Tree
from ..optimizers._base import FunctionalNode
from ..optimizers._base import TerminalNode
from ..optimizers._base import EphemeralConstantNode
from .generators import growing_method
from .generators import sattolo_shuffle
from .transformations import protect_norm
from .transformations import common_region
import random


min_value = np.finfo(np.float64).min
max_value = np.finfo(np.float64).max


################################## MUTATUIONS ##################################
# genetic algorithm
def flip_mutation(individ, proba):
    mask = np.random.random(size=individ.shape) < proba
    individ[mask] = 1 - individ[mask]
    return individ


# differential evolution
def best_1(current, population, F_value):
    best = population[-1]
    r1, r2 = np.random.choice(range(len(population)), size=2, replace=False)
    return best + F_value*(population[r1] - population[r2])


def rand_1(current, population, F_value):
    r1, r2, r3 = np.random.choice(
        range(len(population)), size=3, replace=False)
    return population[r3] + F_value*(population[r1] - population[r2])


def rand_to_best1(current, population, F_value):
    best = population[-1]
    r1, r2, r3 = np.random.choice(
        range(len(population)), size=3, replace=False)
    return population[r1] + F_value*(
        best - population[r1]) + F_value*(population[r2] - population[r3])


def current_to_best_1(current, population, F_value):
    best = population[-1]
    r1, r2 = np.random.choice(range(len(population)), size=2, replace=False)
    return current + F_value*(best - current) +\
        F_value*(population[r1] - population[r2])


def best_2(current, population, F_value):
    best = population[-1]
    r1, r2, r3, r4 = np.random.choice(
        range(len(population)), size=4, replace=False)
    return best + F_value*(population[r1] - population[r2]) +\
        F_value*(population[r3] - population[r4])


def rand_2(current, population, F_value):
    r1, r2, r3, r4, r5 = np.random.choice(
        range(len(population)), size=5, replace=False)
    return population[r5] + F_value*(population[r1] - population[r2]) +\
        F_value*(population[r3] - population[r4])


def current_to_pbest_1(current, population, F_value):
    p_min = 2/len(population)
    p_i = np.random.uniform(p_min, 0.2)

    value = int(p_i*len(population))
    pbest = population[-value:]
    p_best_ind = np.random.randint(0, len(pbest))

    best = pbest[p_best_ind]

    r1, r2 = np.random.choice(range(len(population)), size=2, replace=False)
    return current + F_value*(best - current) +\
        F_value*(population[r1] - population[r2])


def current_to_rand_1(current, population, F_value):
    r1, r2, r3 = np.random.choice(
        range(len(population)), size=3, replace=False)
    return population[r1] + F_value*(population[r3] - current) +\
        F_value*(population[r1] - population[r2])


# genetic propramming
def point_mutation(some_tree, uniset,
                   proba, max_level):
    to_return = some_tree.copy()
    if random.random() < proba:
    # if np.random.random() < proba:
        i = random.randrange(len(to_return))
        # i = np.random.randint(0, len(to_return))
        if type(to_return.nodes[i]) != FunctionalNode:
            new_node = uniset.random_terminal_or_ephemeral()
        else:
            n_args = to_return.nodes[i].n_args
            new_node = uniset.random_functional(n_args)
        to_return.nodes[i] = new_node

    return to_return


def ephemeral_mutation(some_tree, uniset,
                       proba, max_level):
    to_return = some_tree.copy()
    if np.random.random() < proba:
        indexes = [i for i, nodes in enumerate(to_return.nodes)
                   if type(nodes) == EphemeralConstantNode]
        if len(indexes) > 0:
            i = random.choice(indexes)
            new_node = uniset.random_ephemeral()
            to_return.nodes[i] = new_node

    return to_return


def ephemeral_gauss_mutation(some_tree, uniset,
                             proba, max_level):
    to_return = some_tree.copy()
    if np.random.random() < proba:
        indexes = [i for i, nodes in enumerate(to_return.nodes)
                   if type(nodes) == EphemeralConstantNode]
        if len(indexes) > 0:
            i = random.choice(indexes)
            value = to_return.nodes[i].value
            gauss = np.random.normal(1, 0.015)
            new_node = uniset.random_ephemeral()
            new_value = value*gauss

            new_node.value = new_value
            new_node.name = str(new_value)
            new_node.sign = str(new_value)

            to_return.nodes[i] = new_node

    return to_return


def terminal_mutation(some_tree, uniset,
                      proba, max_level):
    to_return = some_tree.copy()
    if np.random.random() < proba:
        indexes = [i for i, nodes in enumerate(to_return.nodes)
                   if type(nodes) == TerminalNode]
        if len(indexes) > 0:
            i = random.choice(indexes)
            new_node = uniset.random_terminal()
            to_return.nodes[i] = new_node

    return to_return


def growing_mutation(some_tree, uniset,
                     proba, max_level):
    to_return = some_tree.copy()
    if random.random() < proba:
    # if np.random.random() < proba:
        # i = np.random.randint(0, len(to_return))
        i = random.randrange(len(to_return))
        left, right = to_return.subtree(i)
        max_level_i = max_level - to_return.levels[left:right][0]
        new_tree = growing_method(uniset, max_level_i)
        to_return = to_return.concat(i, new_tree)

    return to_return


def swap_mutation(some_tree, uniset,
                  proba, max_level):
    to_return = some_tree.copy()
    if np.random.random() < proba:
        indexes = [i for i, nodes in enumerate(
            to_return.nodes) if nodes.n_args > 1]
        if len(indexes) > 0:
            i = random.choice(indexes)
            args_id = to_return.get_args_id(i)
            new_arg_id = args_id.copy()
            sattolo_shuffle(new_arg_id)

            for old_j, new_j in zip(args_id, new_arg_id):
                subtree = some_tree.subtree(old_j, return_class=True)
                to_return = to_return.concat(new_j, subtree)

            to_return.levels = to_return.get_levels()

    return to_return


def shrink_mutation(some_tree, uniset,
                    proba, max_level):
    to_return = some_tree.copy()
    if len(to_return) > 2:
        if np.random.random() < proba:
            indexes = [i for i, nodes in enumerate(to_return.nodes)
                       if nodes.n_args > 0]
            if len(indexes) > 0:
                i = random.choice(indexes)
                args_id = to_return.get_args_id(i)
                choosen = random.choice(args_id)
                to_return = to_return.concat(i, some_tree.subtree(
                    choosen, return_class=True))

    return to_return


################################## CROSSOVERS ##################################
# genetic algorithm
def empty_crossover(individs, *args):
    return individs[0]


def one_point_crossover(individs, fitness, rank):
    cross_point = np.random.randint(0, len(individs[0]))
    if np.random.random() > 0.5:
        offspring = individs[0].copy()
        offspring[:cross_point] = individs[1][:cross_point].copy()
    else:
        offspring = individs[1].copy()
        offspring[:cross_point] = individs[0][:cross_point].copy()
    return offspring


def two_point_crossover(individs, fitness, rank):
    c_point_1, c_point_2 = np.sort(np.random.choice(range(len(individs[0])),
                                                    size=2,
                                                    replace=False))
    if np.random.random() > 0.5:
        offspring = individs[0].copy()
        offspring[c_point_1:c_point_2] = \
            individs[1][c_point_1:c_point_2].copy()
    else:
        offspring = individs[1].copy()
        offspring[c_point_1:c_point_2] = \
            individs[0][c_point_1:c_point_2].copy()

    return offspring


def uniform_crossover(individs, fitness, rank):
    choosen = np.random.choice(range(len(individs)),
                               size=len(individs[0]))
    diag = range(len(individs[0]))
    return individs[choosen, diag]


def uniform_prop_crossover(individs, fitness, rank):
    probability = protect_norm(fitness)
    choosen = np.random.choice(range(len(individs)),
                               size=len(individs[0]), p=probability)

    diag = range(len(individs[0]))
    return individs[choosen, diag]


def uniform_rank_crossover(individs, fitness, rank):
    probability = protect_norm(rank)
    choosen = np.random.choice(range(len(individs)),
                               size=len(individs[0]), p=probability)

    diag = range(len(individs[0]))
    return individs[choosen, diag]


def uniform_tour_crossover(individs, fitness, rank):
    tournament = np.random.choice(range(len(individs)), 2*len(individs[0]))
    tournament = tournament.reshape(-1, 2)

    choosen = np.argmin(fitness[tournament], axis=1)
    diag = range(len(individs[0]))
    return individs[choosen, diag]


# differential evolution
def binomial(individ, mutant, CR):
    individ = individ.copy()
    j = np.random.choice(range(len(individ)), size=1)[0]
    mask_random = np.random.random(len(individ)) <= CR
    mask_j = np.arange(len(individ)) == j
    mask_union = mask_random | mask_j
    individ[mask_union] = mutant[mask_union].copy()
    return individ


# genetic propramming
def standart_crossover(individs, fitness, rank, max_level):
    individ_1 = individs[0].copy()
    individ_2 = individs[1].copy()
    first_point = random.randrange(len(individ_1))
    second_point = random.randrange(len(individ_2))
    # first_point = np.random.randint(0,  len(individ_1))
    # second_point = np.random.randint(0,  len(individ_2))

    if random.random() < 0.5:
        # if np.random.random() < 0.5:
        first_subtree = individ_1.subtree(first_point, return_class=True)
        offspring = individ_2.concat(second_point, first_subtree)
        if offspring.get_max_level() > max_level:
            offspring = individ_2
    else:
        second_subtree = individ_2.subtree(second_point, return_class=True)
        offspring = individ_1.concat(first_point, second_subtree)
        if offspring.get_max_level() > max_level:
            offspring = individ_1
    return offspring


def one_point_crossoverGP(individs, fitness, rank, max_level):
    individ_1 = individs[0]
    individ_2 = individs[1]
    common_indexes, _ = common_region([individ_1, individ_2])

    # point = np.random.randint(0,  len(common_indexes[0]))
    point = random.randrange(len(common_indexes[0]))
    first_point = common_indexes[0][point]
    second_point = common_indexes[1][point]
    if random.random() < 0.5:
    # if np.random.random() < 0.5:
        first_subtree = individ_1.subtree(first_point, return_class=True)
        offspring = individ_2.concat(second_point, first_subtree)
    else:
        second_subtree = individ_2.subtree(second_point, return_class=True)
        offspring = individ_1.concat(first_point, second_subtree)
    return offspring


def uniform_crossoverGPc(individs, fitness, rank, max_level):
    '''Poli, Riccardo & Langdon, W.. (2001). On the Search
    Properties of Different Crossover Operators in Genetic Programming. '''
    new_nodes = []
    individ_1 = individs[0]
    individ_2 = individs[1]
    common_indexes, border = common_region([individ_1, individ_2])

    for i in range(len(common_indexes[0])):
        if common_indexes[0][i] in border[0]:
            if np.random.random() < 0.5:
                id_ = common_indexes[0][i]
                left, right = individ_1.subtree(index=id_)
                new_nodes.extend(individ_1.nodes[left:right])
            else:
                id_ = common_indexes[1][i]
                left, right = individ_2.subtree(index=id_)
                new_nodes.extend(individ_2.nodes[left:right])
        else:
            if np.random.random() < 0.5:
                id_ = common_indexes[0][i]
                new_nodes.append(individ_1.nodes[id_])
            else:
                id_ = common_indexes[1][i]
                new_nodes.append(individ_2.nodes[id_])
    to_return = Tree(new_nodes.copy())
    to_return.levels = to_return.get_levels()
    return to_return


def uniform_crossoverGP(individs, fitness, rank, max_level):
    range_ = range(len(individs))
    to_return = Tree([])
    common, border = common_region(individs)
    for i, common_0_i in enumerate(common[0]):
        j = np.random.choice(range_, 1)[0]
        id_ = common[j][i]
        to_return.nodes.append(individs[j].nodes[id_])
        if common_0_i in border[0]:
            left_subtrees = []
            right_subtrees = []

            for k, tree_k in enumerate(individs):
                inner_id = common[k][i]
                args_id = tree_k.get_args_id(inner_id)

                n_args = tree_k.nodes[inner_id].n_args
                if n_args == 1:
                    subtree = tree_k.subtree(args_id[0], return_class=True)
                    left_subtrees.append(subtree)
                    right_subtrees.append(subtree)
                elif n_args == 2:
                    subtree_l = tree_k.subtree(args_id[0], return_class=True)
                    subtree_r = tree_k.subtree(args_id[1], return_class=True)
                    left_subtrees.append(subtree_l)
                    right_subtrees.append(subtree_r)

            n_args = individs[j].nodes[id_].n_args
            if n_args == 1:
                choosen = random.choices(
                    left_subtrees + right_subtrees, k=1)[0]
                to_return.nodes.extend(choosen.nodes)
            elif n_args == 2:
                choosen_l = random.choices(left_subtrees, k=1)[0]
                to_return.nodes.extend(choosen_l.nodes)

                choosen_r = random.choices(right_subtrees, k=1)[0]
                to_return.nodes.extend(choosen_r.nodes)

    to_return = to_return.copy()
    to_return.levels = to_return.get_levels()
    return to_return


def uniform_crossoverGP_prop(individs, fitness, rank, max_level):
    range_ = range(len(individs))
    probability = protect_norm(fitness)
    to_return = Tree([])
    common, border = common_region(individs)
    for i, common_0_i in enumerate(common[0]):
        j = np.random.choice(range_, 1, p=probability)[0]
        id_ = common[j][i]
        to_return.nodes.append(individs[j].nodes[id_])
        if common_0_i in border[0]:
            left_subtrees = []
            right_subtrees = []
            left_fitness = []
            right_fitness = []

            for k, tree_k in enumerate(individs):
                inner_id = common[k][i]
                args_id = tree_k.get_args_id(inner_id)
                n_args = tree_k.nodes[inner_id].n_args
                if n_args == 1:
                    subtree = tree_k.subtree(args_id[0], return_class=True)
                    left_subtrees.append(subtree)
                    right_subtrees.append(subtree)
                    left_fitness.append(fitness[k])
                    right_fitness.append(fitness[k])
                elif n_args == 2:
                    subtree_l = tree_k.subtree(args_id[0], return_class=True)
                    subtree_r = tree_k.subtree(args_id[1], return_class=True)
                    left_subtrees.append(subtree_l)
                    right_subtrees.append(subtree_r)
                    left_fitness.append(fitness[k])
                    right_fitness.append(fitness[k])

            n_args = individs[j].nodes[id_].n_args
            if n_args == 1:
                fitness_i = np.array(left_fitness + right_fitness)
                proba = protect_norm(fitness_i)
                choosen = random.choices(
                    left_subtrees + right_subtrees, weights=proba, k=1)[0]
                to_return.nodes.extend(choosen.nodes)
            elif n_args == 2:
                fitness_l = np.array(left_fitness)
                fitness_r = np.array(right_fitness)
                proba_l = protect_norm(fitness_l)
                proba_r = protect_norm(fitness_r)

                choosen_l = random.choices(
                    left_subtrees, weights=proba_l, k=1)[0]
                to_return.nodes.extend(choosen_l.nodes)

                choosen_r = random.choices(
                    right_subtrees, weights=proba_r, k=1)[0]
                to_return.nodes.extend(choosen_r.nodes)

    to_return = to_return.copy()
    to_return.levels = to_return.get_levels()
    return to_return


def uniform_crossoverGP_rank(individs, fitness, rank, max_level):
    range_ = range(len(individs))
    probability = protect_norm(rank)
    to_return = Tree([])
    common, border = common_region(individs)
    for i, common_0_i in enumerate(common[0]):
        j = random.choices(range_, weights=probability)[0]
        # j = np.random.choice(range_, 1, p=probability)[0]
        id_ = common[j][i]
        to_return.nodes.append(individs[j].nodes[id_])
        if common_0_i in border[0]:
            left_subtrees = []
            right_subtrees = []
            left_fitness = []
            right_fitness = []

            for k, tree_k in enumerate(individs):
                inner_id = common[k][i]
                args_id = tree_k.get_args_id(inner_id)
                n_args = tree_k.nodes[inner_id].n_args
                if n_args == 1:
                    subtree = tree_k.subtree(args_id[0], return_class=True)
                    left_subtrees.append(subtree)
                    right_subtrees.append(subtree)
                    left_fitness.append(rank[k])
                    right_fitness.append(rank[k])
                elif n_args == 2:
                    subtree_l = tree_k.subtree(args_id[0], return_class=True)
                    subtree_r = tree_k.subtree(args_id[1], return_class=True)
                    left_subtrees.append(subtree_l)
                    right_subtrees.append(subtree_r)
                    left_fitness.append(rank[k])
                    right_fitness.append(rank[k])

            n_args = individs[j].nodes[id_].n_args
            if n_args == 1:
                fitness_i = np.array(left_fitness + right_fitness)
                proba = protect_norm(fitness_i)
                choosen = random.choices(
                    left_subtrees + right_subtrees, weights=proba, k=1)[0]
                to_return.nodes.extend(choosen.nodes)
            elif n_args == 2:
                fitness_l = np.array(left_fitness)
                fitness_r = np.array(right_fitness)
                proba_l = protect_norm(fitness_l)
                proba_r = protect_norm(fitness_r)

                choosen_l = random.choices(
                    left_subtrees, weights=proba_l, k=1)[0]
                to_return.nodes.extend(choosen_l.nodes)

                choosen_r = random.choices(
                    right_subtrees, weights=proba_r, k=1)[0]
                to_return.nodes.extend(choosen_r.nodes)

    to_return = to_return.copy()
    to_return.levels = to_return.get_levels()
    return to_return


def uniform_crossoverGP_tour(individs, fitness, rank, max_level):
    range_ = range(len(individs))
    to_return = Tree([])
    common, border = common_region(individs)
    for i, common_0_i in enumerate(common[0]):

        j = tournament_selection(fitness, rank, 2, 1)[0]
        id_ = common[j][i]
        to_return.nodes.append(individs[j].nodes[id_])
        if common_0_i in border[0]:
            left_subtrees = []
            right_subtrees = []
            left_fitness = []
            right_fitness = []

            for k, tree_k in enumerate(individs):
                inner_id = common[k][i]
                args_id = tree_k.get_args_id(inner_id)
                n_args = tree_k.nodes[inner_id].n_args
                if n_args == 1:
                    subtree = tree_k.subtree(args_id[0], return_class=True)
                    left_subtrees.append(subtree)
                    right_subtrees.append(subtree)
                    left_fitness.append(rank[k])
                    right_fitness.append(rank[k])
                elif n_args == 2:
                    subtree_l = tree_k.subtree(args_id[0], return_class=True)
                    subtree_r = tree_k.subtree(args_id[1], return_class=True)
                    left_subtrees.append(subtree_l)
                    right_subtrees.append(subtree_r)
                    left_fitness.append(rank[k])
                    right_fitness.append(rank[k])

            n_args = individs[j].nodes[id_].n_args
            if n_args == 1:
                fitness_i = np.array(left_fitness + right_fitness)
                proba = protect_norm(fitness_i)
                choosen = random.choices(
                    left_subtrees + right_subtrees, weights=proba, k=1)[0]
                to_return.nodes.extend(choosen.nodes)
            elif n_args == 2:
                fitness_l = np.array(left_fitness)
                fitness_r = np.array(right_fitness)
                proba_l = protect_norm(fitness_l)
                proba_r = protect_norm(fitness_r)

                choosen_l = random.choices(
                    left_subtrees, weights=proba_l, k=1)[0]
                to_return.nodes.extend(choosen_l.nodes)

                choosen_r = random.choices(
                    right_subtrees, weights=proba_r, k=1)[0]
                to_return.nodes.extend(choosen_r.nodes)

    to_return = to_return.copy()
    to_return.levels = to_return.get_levels()
    return to_return


################################## SELECTIONS ##################################
# genetic algorithm
def proportional_selection(fitness, ranks,
                           tour_size, quantity):
    probability = fitness/fitness.sum()
    # choosen = np.random.choice(range(len(fitness)),
    #                            size=quantity, p=probability)
    choosen = random.choices(range(len(fitness)),
                                k=quantity, weights=probability)
    return choosen


def rank_selection(fitness, ranks,
                   tour_size, quantity):
    probability = ranks/np.sum(ranks)
    # choosen = np.random.choice(range(len(fitness)),
    #                            size=quantity, p=probability)
    choosen = random.choices(range(len(fitness)),
                                k=quantity, weights=probability)
    return choosen


def tournament_selection(fitness, ranks,
                         tour_size, quantity):
    # tournament = np.random.choice(
    #     range(len(fitness)), tour_size*quantity)
    tournament = random.choices(range(len(fitness)),
                                k=tour_size*quantity)
    tournament = np.array(tournament).reshape(-1, tour_size)
    # tournament = tournament.reshape(-1, tour_size)
    max_fit_id = np.argmax(fitness[tournament], axis=1)
    choosen = np.diag(tournament[:, max_fit_id])
    return choosen


##################################### MATH #####################################
class Operator:
    def write(self, *args):
        return self.formula.format(*args)


class Cos(Operator):
    def __init__(self):
        self.formula = 'cos({})'
        self.__name__ = 'cos'
        self.sign = 'cos'

    def __call__(self, x):
        return np.cos(x)


class Sin(Operator):
    def __init__(self):
        self.formula = 'sin({})'
        self.__name__ = 'sin'
        self.sign = 'sin'

    def __call__(self, x):
        return np.sin(x)


class Add(Operator):
    def __init__(self):
        self.formula = '({} + {})'
        self.__name__ = 'add'
        self.sign = '+'

    def __call__(self, x, y):
        return x + y

class Sub(Operator):
    def __init__(self):
        self.formula = '({} - {})'
        self.__name__ = 'sub'
        self.sign = '-'

    def __call__(self, x, y):
        return x - y


class Neg(Operator):
    def __init__(self):
        self.formula = '-{}'
        self.__name__ = 'neg'
        self.sign = '-'

    def __call__(self, x):
        return -x



class Mul(Operator):
    def __init__(self):
        self.formula = '({} * {})'
        self.__name__ = 'mul'
        self.sign = '*'

    def __call__(self, x, y):
        return x * y


class Pow2(Operator):
    def __init__(self):
        self.formula = '({}**2)'
        self.__name__ = 'pow2'
        self.sign = '**2'

    def __call__(self, x):
        res = x**2
        return np.clip(res, min_value, max_value)


# class Pow3(Operator):
#     def __init__(self):
#         self.formula = '({}**3)'
#         self.__name__ = 'pow3'
#         self.sign = '**3'

#     def __call__(self, x):
#         res = x**3
#         return np.clip(res, min_value, max_value)

class Div(Operator):
    def __init__(self):
        self.formula = '({}/{})'
        self.__name__ = 'div'
        self.sign = '/'

    def __call__(self, x, y):
        if type(y) == np.ndarray:
            res = np.divide(x, y, out=np.ones_like(
                y, dtype=np.float64), where=y != 0)
        else:
            if y == 0:
                res = 0
            else:
                res = x/y
        return np.clip(res, min_value, max_value)

# class FloorDiv(Operator):
#     def __init__(self):
#         self.formula = '({}//{})'
#         self.__name__ = 'floor_div'
#         self.sign = '//'

#     def __call__(self, x, y):
#         if type(y) == np.ndarray:
#             res = np.floor_divide(x, y, out=np.ones_like(
#                 y, dtype=np.float64), where=y != 0)
#         else:
#             if y == 0:
#                 res = 0
#             else:
#                 res = x//y
#         return np.clip(res, min_value, max_value)`


class Inv(Operator):
    def __init__(self):
        self.formula = '(1/{})'
        self.__name__ = 'Ind'
        self.sign = '1/'

    def __call__(self, y):
        if type(y) == np.ndarray:
            res = np.divide(1, y, out=np.ones_like(
                y, dtype=np.float64), where=y != 0)
        else:
            if y == 0:
                res = 1
            else:
                res = 1/y
        return np.clip(res, min_value, max_value)
        # return res



class LogAbs(Operator):
    def __init__(self):
        self.formula = 'log(abs({}))'
        self.__name__ = 'log(abs)'
        self.sign = 'log(abs)'

    def __call__(self, y):
        y_ = np.abs(y)
        if type(y_) == np.ndarray:
            res = np.log(y_, out=np.ones_like(
                y_, dtype=np.float64), where=y_ != 0)
        else:
            if y_ == 0:
                res = 1
            else:
                res = np.log(y_)
        return np.clip(res, min_value, max_value)


class Exp(Operator):
    def __init__(self):
        self.formula = 'exp({})'
        self.__name__ = 'exp'
        self.sign = 'exp'

    def __call__(self, x):
        return np.clip(np.exp(x), min_value, max_value)


class Mul3(Operator):
    def __init__(self):
        self.formula = '({} * {} * {})'
        self.__name__ = 'mul3'
        self.sign = '*'

    def __call__(self, x, y, z):
        return np.clip(x * y * z, min_value, max_value)


class SqrtAbs(Operator):
    def __init__(self):
        self.formula = 'sqrt(abs({}))'
        self.__name__ = 'sqrt(abs)'
        self.sign = 'sqrt(abs)'

    def __call__(self, x):
        return np.clip(np.sqrt(np.abs(x)), min_value, max_value)


# class Abs(Operator):
#     def __init__(self):
#             self.formula = 'abs({})'
#             self.__name__ = 'abs()'
#             self.sign = 'abs()'

#     def __call__(self, x):
#         return np.clip(np.abs(x), min_value, max_value)



import numpy as np
from ._base import TheFittest
from ._base import LastBest
from functools import partial
from ._differentialevolution import DifferentialEvolution
from ..tools.operators import binomial
from ..tools.operators import rand_1
from ..tools.operators import current_to_best_1
from ..tools.transformations import numpy_group_by


class StaticticSaDE:
    def __init__(self, mode='quick'):
        self.mode = mode
        self.population_g = np.array([])
        self.fitness = np.array([])
        self.m_proba = dict()
        self.CRm = np.array([], dtype=float)

    def append_arr(self, arr_to, arr_from):
        shape_to = (-1, arr_from.shape[0], arr_from.shape[1])
        shape_from = (1, arr_from.shape[0], arr_from.shape[1])
        result = np.vstack([arr_to.reshape(shape_to),
                            arr_from.copy().reshape(shape_from)])
        return result

    def update(self,
               population_g_i,
               fitness_i,
               m_proba_i, CRm_i):
        if self.mode == 'quick':
            self.fitness = np.append(self.fitness, np.max(fitness_i))
            if not len(self.m_proba):
                for key, value in m_proba_i.items():
                    self.m_proba[key] = np.array(value)
            else:
                for key, value in m_proba_i.items():
                    self.m_proba[key] = np.append(
                        self.m_proba[key], np.array(value))
            self.CRm = np.append(self.CRm, CRm_i)
        elif self.mode == 'full':
            self.fitness = np.append(self.fitness, np.max(fitness_i))
            self.population_g = self.append_arr(self.population_g,
                                                population_g_i)
            if not len(self.m_proba):
                for key, value in m_proba_i.items():
                    self.m_proba[key] = np.array(value)
            else:
                for key, value in m_proba_i.items():
                    self.m_proba[key] = np.append(
                        self.m_proba[key], np.array(value))
            self.CRm = np.append(self.CRm, CRm_i)
        else:
            raise ValueError('the "mode" must be either "quick" or "full"')
        return self


class SaDE2005(DifferentialEvolution):
    '''Qin, Kai & Suganthan, Ponnuthurai. (2005). Self-adaptive differential evolution
    algorithm for numerical optimization. 2005 IEEE Congress on Evolutionary Computation,
    IEEE CEC 2005. Proceedings. 2. 1785-1791. 10.1109/CEC.2005.1554904'''

    def __init__(self,
                 fitness_function,
                 genotype_to_phenotype,
                 iters,
                 pop_size,
                 left,
                 right,
                 optimal_value = None,
                 termination_error_value = 0.,
                 no_increase_num = None,
                 minimization = False,
                 show_progress_each = None,
                 keep_history = None):
        DifferentialEvolution.__init__(
            self,
            fitness_function=fitness_function,
            genotype_to_phenotype=genotype_to_phenotype,
            iters=iters,
            pop_size=pop_size,
            left=left,
            right=right,
            optimal_value=optimal_value,
            termination_error_value=termination_error_value,
            no_increase_num=no_increase_num,
            minimization=minimization,
            show_progress_each=show_progress_each,
            keep_history=keep_history)

        self.m_sets = {'rand_1': rand_1,
                       'current_to_best_1': current_to_best_1}
        self.m_sets = dict(sorted(self.m_sets.items()))
        self.m_sets_keys = list(self.m_sets.keys())

        self.thefittest: TheFittest
        self.stats: StaticticSaDE
        self.m_learning_period: int
        self.CR_update_timer: int
        self.CR_m_learning_period: int
        self.threshold: int

        self.Fm = 0.5
        self.F_sigma = 0.3
        self.CR_sigma = 0.1

        self.set_strategy()

    def set_strategy(self, m_learning_period_param = 50,
                     CR_update_timer_param = 5,
                     CR_m_learning_period_param = 25,
                     threshold_params = 0.1,
                     Fm_param = 0.5,
                     F_sigma_param = 0.3,
                     CR_sigma_param = 0.1,
                     elitism_param = True,
                     initial_population = None):
        self.update_pool()
        self.m_learning_period = m_learning_period_param
        self.CR_update_timer = CR_update_timer_param
        self.CR_m_learning_period = CR_m_learning_period_param
        self.threshold = threshold_params
        self.Fm = Fm_param
        self.F_sigma = F_sigma_param
        self.CR_sigma = CR_sigma_param
        self.elitism = elitism_param
        self.initial_population = initial_population
        return self

    def mutation_and_crossover(self, popuation_g, individ_g,
                               mutation_type, CR_i):
        F_i = np.random.normal(self.Fm, self.F_sigma)
        mutant = self.m_sets[mutation_type](individ_g, popuation_g, F_i)

        mutant_cr_g = binomial(individ_g, mutant, CR_i)
        mutant_cr_g = self.bounds_control(mutant_cr_g)
        return mutant_cr_g

    def evaluate_and_selection(self, mutant_cr_g, population_g, population_ph, fitness):
        offspring_g = population_g.copy()
        offspring_ph = population_ph.copy()
        offspring_fit = fitness.copy()

        mutant_cr_ph = self.genotype_to_phenotype(mutant_cr_g)
        mutant_cr_fit = self.evaluate(mutant_cr_ph)
        mask = mutant_cr_fit >= fitness
        offspring_g[mask] = mutant_cr_g[mask]
        offspring_ph[mask] = mutant_cr_ph[mask]
        offspring_fit[mask] = mutant_cr_fit[mask]
        return (offspring_g, offspring_ph, offspring_fit, mask)

    def choice_operators(self, proba_dict):
        operators = list(proba_dict.keys())
        proba = list(proba_dict.values())
        return np.random.choice(list(operators), self.pop_size, p=proba)

    def succeses_to_ns(self, succeses):
        return np.sum(succeses)

    def succeses_to_nf(self, succeses):
        return np.sum(~succeses)

    def update_ns_nf(self, operators, succeses, ns_i, nf_i):
        grouped = dict(zip(*numpy_group_by(group=succeses, by=operators)))

        for key in self.m_sets.keys():
            if key in grouped.keys():
                ns_i[key] += self.succeses_to_ns(grouped[key])
                nf_i[key] += self.succeses_to_nf(grouped[key])
        return ns_i, nf_i

    def update_proba(self, ns_i, nf_i):
        up = ns_i['rand_1']*(ns_i['current_to_best_1'] +
                             nf_i['current_to_best_1'])
        down = ns_i['current_to_best_1']*(ns_i['rand_1'] + nf_i['rand_1']) + up

        p1 = up/down
        return {'rand_1': p1, 'current_to_best_1': 1 - p1}

    def generate_CR(self, size, CRm):
        value = np.random.normal(CRm, self.CR_sigma, size)
        value = np.clip(value, 1e-6, 1)
        return value

    def if_period_ended(self, i, period):
        return i % period == 0 and i != 0

    def fit(self):
        z_m = len(self.m_sets)
        m_proba = dict(zip(self.m_sets_keys, np.full(z_m, 1/z_m)))
        CRm = 0.5
        CR_i = self.generate_CR(self.pop_size, CRm)

        population_g = self.generate_init_pop()
        population_ph = self.genotype_to_phenotype(population_g)
        fitness = self.evaluate(population_ph)

        argsort = np.argsort(fitness)
        population_g = population_g[argsort]
        population_ph = population_ph[argsort]
        fitness = fitness[argsort]

        self.thefittest = TheFittest().update(population_g,
                                              population_ph,
                                              fitness)
        lastbest = LastBest().update(self.thefittest.fitness)
        if self.keep_history is not None:
            self.stats = StaticticSaDE(
                mode=self.keep_history).update(population_g,
                                               fitness,
                                               m_proba,
                                               CRm)

        ns = dict(zip(self.m_sets_keys, np.zeros(z_m, dtype=int)))
        nf = dict(zip(self.m_sets_keys, np.zeros(z_m, dtype=int)))
        CR_s_pool = np.array([], dtype=float)
        for i in range(self.iters-1):
            self.show_progress(i)
            if self.termitation_check(lastbest.no_increase):
                break
            else:
                m_operators = self.choice_operators(m_proba)

                partial_mut_and_cross = partial(self.mutation_and_crossover,
                                                population_g)
                mutant_cr_g = np.array(list(map(partial_mut_and_cross,
                                                population_g,
                                                m_operators, CR_i)))

                stack = self.evaluate_and_selection(mutant_cr_g,
                                                    population_g,
                                                    population_ph,
                                                    fitness)

                population_g = stack[0]
                population_ph = stack[1]
                fitness = stack[2]

                if self.elitism:
                    population_g[-1], population_ph[-1], fitness[-1] = self.thefittest.get()
                argsort = np.argsort(fitness)
                population_g = population_g[argsort]
                population_ph = population_ph[argsort]
                fitness = fitness[argsort]

                self.thefittest.update(population_g, population_ph, fitness)
                lastbest.update(self.thefittest.fitness)
                if self.keep_history is not None:
                    self.stats.update(population_g,
                                      fitness,
                                      m_proba,
                                      CRm)
                successes = stack[-1]
                ns, nf = self.update_ns_nf(m_operators, successes, ns, nf)
                CR_s_pool = np.append(CR_s_pool, CR_i[successes])

                if self.if_period_ended(i, self.m_learning_period):
                    m_proba = self.update_proba(ns, nf)
                    ns = dict(zip(self.m_sets_keys, np.zeros(z_m, dtype=int)))
                    nf = dict(zip(self.m_sets_keys, np.zeros(z_m, dtype=int)))

                if self.if_period_ended(i, self.CR_update_timer):
                    CR_i = self.generate_CR(self.pop_size, CRm)
                    if self.if_period_ended(i, self.CR_m_learning_period):
                        if len(CR_s_pool):
                            CRm = np.mean(CR_s_pool)
                        CR_s_pool = np.array([], dtype=float)
        return self



import numpy as np
from typing import Any
import random
from inspect import signature


class LastBest:
    def __init__(self):
        self.value = np.nan
        self.no_increase = 0

    def update(self, current_value: float):
        if self.value == current_value:
            self.no_increase += 1
        else:
            self.no_increase = 0
            self.value = current_value
        return self


class TheFittest:
    def __init__(self):
        self.genotype: Any
        self.phenotype: Any
        self.fitness = -np.inf

    def update(self, population_g, population_ph, fitness):
        temp_best_id = np.argmax(fitness)
        temp_best_fitness = fitness[temp_best_id].copy()
        if temp_best_fitness > self.fitness:
            self.genotype = population_g[temp_best_id].copy()
            self.phenotype = population_ph[temp_best_id].copy()
            self.fitness = temp_best_fitness.copy()

        return self

    def get(self):
        return self.genotype.copy(), self.phenotype.copy(), self.fitness.copy()


class EvolutionaryAlgorithm:
    def __init__(self,
                 fitness_function,
                 genotype_to_phenotype,
                 iters,
                 pop_size,
                 optimal_value = None,
                 termination_error_value = 0.,
                 no_increase_num = None,
                 minimization = False,
                 show_progress_each = None,
                 keep_history = None):
        self.fitness_function = fitness_function
        self.genotype_to_phenotype = genotype_to_phenotype
        self.iters = iters
        self.pop_size = pop_size
        self.no_increase_num = no_increase_num
        self.show_progress_each = show_progress_each
        self.keep_history = keep_history

        self.sign = -1 if minimization else 1

        if optimal_value is not None:
            self.aim = self.sign*optimal_value - termination_error_value
        else:
            self.aim = np.inf

        self.calls = 0

    def evaluate(self, population_ph):
        self.calls += len(population_ph)
        return self.sign*self.fitness_function(population_ph)

    def show_progress(self, i):
        if (self.show_progress_each is not None) and (i % self.show_progress_each == 0):
            print(f'{i} iteration with fitness = {self.thefittest.fitness}')

    def termitation_check(self, no_increase):
        return (self.thefittest.fitness >= self.aim) or (no_increase == self.no_increase_num)

    def get_remains_calls(self):
        return (self.pop_size + (self.iters-1)*(self.pop_size-1)) - self.calls


class Tree:
    def __init__(self, nodes):
        self.nodes = nodes
        self.levels = []

    def __len__(self):
        return len(self.nodes)

    def __str__(self):
        reverse_nodes = self.nodes[::-1].copy()
        pack = []
        for node in reverse_nodes:
            args = []
            for _ in range(node.n_args):
                args.append(pack.pop())
            if type(node) != FunctionalNode:
                pack.append(node.name)
            else:
                pack.append(node.value.write(*args))
        return pack[0]

    def __eq__(self, other):
        if len(self) != len(other):
            return False
        else:
            for node_1, node_2 in zip(self.nodes, other.nodes):
                if np.all(node_1.value != node_2.value):
                    return False
        return True

    def subtree(self, index, return_class=False):
        n_index = index + 1
        possible_steps = self.nodes[index].n_args
        while possible_steps:
            possible_steps += self.nodes[n_index].n_args - 1
            n_index += 1
        if return_class:
            new_tree = Tree(self.nodes[index:n_index].copy())
            return new_tree
        return index, n_index

    def compile(self):
        pack = []
        for node in reversed(self.nodes):
            args = []
            for _ in range(node.n_args):
                args.append(pack.pop())
            if type(node) != FunctionalNode:
                pack.append(node.value)
            else:
                pack.append(node.value(*args))
        return pack[0]

    def concat(self, index, some_tree):
        left, right = self.subtree(index)
        levels = some_tree.get_levels(origin=self.levels[left])

        new_nodes = self.nodes[:left]
        new_levels = self.levels[:left]

        new_nodes += some_tree.nodes
        new_levels += levels

        new_nodes += self.nodes[right:]
        new_levels += self.levels[right:]
        to_return = Tree(new_nodes.copy())
        to_return.levels = new_levels.copy()
        return to_return

    def copy(self):
        tree = Tree(self.nodes.copy())
        tree.levels = self.levels.copy()
        return tree

    def change_terminals(self, change_list):
        tree_copy = self.copy()
        for i, node in enumerate(tree_copy.nodes):
            if type(node) is TerminalNode:
                for key, value in change_list.items():
                    if node.name == key:
                        tree_copy.nodes[i] = TerminalNode(value=value,
                                                          name=node.name + '_')
        return tree_copy

    def get_max_level(self):
        return np.max(self.levels)

    def get_levels(self, origin=0):
        d_i = origin-1
        s = [1]
        d = [origin-1]
        result_list = []
        for node in self.nodes:
            s[-1] = s[-1] - 1
            if s[-1] == 0:
                s.pop()
                d_i = d.pop() + 1
            else:
                d_i = d[-1] + 1
            result_list.append(d_i)
            if node.n_args > 0:
                s.append(node.n_args)
                d.append(d_i)
        return result_list

    def get_args_id(self, index=0):
        n_args = self.nodes[index].n_args
        args_id = []
        root_level = self.levels[index]
        next_level = root_level + 1
        k = index + 1
        while n_args:
            if self.levels[k] == next_level:
                args_id.append(k)
                n_args = n_args - 1
            k = k + 1
        return args_id

    def get_graph(self, keep_id=False):
        pack = []
        edges = []
        nodes = []
        labels = {}
        for i, node in enumerate(reversed(self.nodes)):
            index = len(self) - i - 1
            if keep_id:
                labels[index] = str(index) + '. ' + node.sign[:6]
            else:
                labels[index] = node.sign[:6]

            nodes.append(index)

            for _ in range(node.n_args):
                edges.append((index, len(self) - pack.pop() - 1))
            pack.append(i)

        edges.reverse()
        nodes.reverse()

        levels = self.levels
        colors = np.zeros(shape=(len(nodes), 4))
        pos = np.zeros(shape=(len(self), 2))
        for i, lvl_i in enumerate(levels):
            total = 0
            cond = lvl_i == np.array(levels)
            h = 1/(1 + np.sum(cond))
            arange = np.arange(len(pos))[cond]

            for j, a_j in enumerate(arange):
                total += h
                pos[a_j][0] = total

            pos[i][1] = -lvl_i

            if type(self.nodes[i]) is FunctionalNode:
                colors[i] = (1, 0.72, 0.43, 1)
            else:
                colors[i] = (0.21, 0.76, 0.56, 1)

        to_return = {'edges': edges,
                     'labels': labels,
                     'nodes': nodes,
                     'pos': pos,
                     'colors': colors}
        return to_return


class Node:
    def __init__(self,
                 value,
                 name,
                 sign,
                 n_args):
        self.value = value
        self.name = name
        self.sign = sign
        self.n_args = n_args

    def __str__(self):
        return str(self.sign)

    def __eq__(self, other):
        return self.name == other.name


class FunctionalNode(Node):
    def __init__(self, value, sign = None):
        Node.__init__(self,
                      value=value,
                      name=value.__name__,
                      sign=sign or value.sign,
                      n_args=len(signature(value).parameters))


class TerminalNode(Node):
    def __init__(self, value, name):
        Node.__init__(self,
                      value=value,
                      name=name,
                      sign=name,
                      n_args=0)


class EphemeralNode():
    def __init__(self, value):
        self.value = value

    def __call__(self):
        return self.value()


class EphemeralConstantNode(Node):
    def __init__(self, generator):
        value = generator()
        Node.__init__(self,
                      value=value,
                      name=str(value),
                      sign=str(value),
                      n_args=0)


'''functional_set = (FunctionalNode, ..., FunctionalNode)
terminal_set = (TerminalNode, ..., TerminalNode)
ephemeral_set = (EphemeralNode, ..., EphemeralNode)'''


class UniversalSet:
    def __init__(self, functional_set, terminal_set, ephemeral_set = []):
        self.functional_set = {'any': functional_set}
        for unit in functional_set:
            n_args = unit.n_args
            if n_args not in self.functional_set:
                self.functional_set[n_args] = [unit]
            else:
                self.functional_set[n_args].append(unit)
        self.union_terminal = list(terminal_set) + list(ephemeral_set)
        self.terminal_set = terminal_set
        self.ephemeral_set = ephemeral_set

    def random_terminal(self):
        choosen = random.choice(self.terminal_set)
        return choosen
    
    def random_ephemeral(self):
        choosen = random.choice(self.ephemeral_set)
        return EphemeralConstantNode(choosen)

    def random_terminal_or_ephemeral(self):
        choosen = random.choice(self.union_terminal)
        if type(choosen) is EphemeralNode:
            to_return = EphemeralConstantNode(choosen)
        else:
            to_return = choosen
        return to_return

    def random_functional(self, n_args='any'):
        choosen = random.choice(self.functional_set[n_args])
        return choosen



import numpy as np
import numpy as np
from typing import Any
from ._base import TheFittest
from ._base import EvolutionaryAlgorithm
from ._base import LastBest
from ..tools.operators import proportional_selection
from ..tools.operators import rank_selection
from ..tools.operators import tournament_selection
from ..tools.operators import one_point_crossoverGP
from ..tools.operators import standart_crossover
from ..tools.operators import uniform_crossoverGPc
from ..tools.operators import uniform_crossoverGP
from ..tools.operators import uniform_crossoverGP_prop
from ..tools.operators import uniform_crossoverGP_rank
from ..tools.operators import uniform_crossoverGP_tour
from ..tools.operators import empty_crossover
from ..tools.operators import point_mutation
from ..tools.operators import ephemeral_mutation
from ..tools.operators import ephemeral_gauss_mutation
from ..tools.operators import terminal_mutation
from ..tools.operators import growing_mutation
from ..tools.operators import swap_mutation
from ..tools.operators import shrink_mutation
from ..tools.generators import half_and_half
from ..tools.transformations import scale_data
from ..tools.transformations import rank_data
from functools import partial


class StatisticsGP:
    def __init__(self, mode='quick'):
        self.mode = mode
        self.fittest = np.array([])
        self.fitness = np.array([])

    def update(self,
               fittest_i,
               fitness_i):
        if self.mode == 'quick':
            self.fitness = np.append(self.fitness, np.max(fitness_i))
        elif self.mode == 'full':
            self.fittest = np.append(self.fittest, fittest_i.copy())
            self.fitness = np.append(self.fitness, np.max(fitness_i))
        else:
            raise ValueError('the "mode" must be either "quick" or "full"')
        return self


class GeneticProgramming(EvolutionaryAlgorithm):
    '''Koza, John R.. “Genetic programming - on the programming of computers by means
    of natural selection.” Complex Adaptive Systems (1993)'''

    def __init__(self,
                 fitness_function,
                 genotype_to_phenotype,
                 uniset,
                 iters,
                 pop_size,
                 optimal_value = None,
                 termination_error_value = 0.,
                 no_increase_num = None,
                 minimization = False,
                 show_progress_each = None,
                 keep_history = None):

        EvolutionaryAlgorithm.__init__(
            self,
            fitness_function=fitness_function,
            genotype_to_phenotype=genotype_to_phenotype,
            iters=iters,
            pop_size=pop_size,
            optimal_value=optimal_value,
            termination_error_value=termination_error_value,
            no_increase_num=no_increase_num,
            minimization=minimization,
            show_progress_each=show_progress_each,
            keep_history=keep_history)

        self.uniset = uniset
        self.thefittest: TheFittest
        self.stats: StatisticsGP
        self.s_pool: dict
        self.c_pool: dict
        self.m_pool: dict
        self.tour_size: int
        self.max_level: int
        self.init_level: int
        self.initial_population: np.ndarray
        self.s_set: tuple
        self.c_set: tuple
        self.m_set: tuple

        self.set_strategy()

    def update_pool(self):
        self.s_pool = {'proportional': (proportional_selection, None),
                       'rank': (rank_selection, None),
                       'tournament_k': (tournament_selection, self.tour_size),
                       'tournament_3': (tournament_selection, 3),
                       'tournament_5': (tournament_selection, 5),
                       'tournament_7': (tournament_selection, 7)}

        self.c_pool = {'empty': (empty_crossover, 1),
                       'standart': (standart_crossover, 2),
                       'one_point': (one_point_crossoverGP, 2),
                       'uniform2c': (uniform_crossoverGPc, 2),
                       'uniform2': (uniform_crossoverGP, 2),
                       'uniform7': (uniform_crossoverGP, 7),
                       'uniformk': (uniform_crossoverGP, self.parents_num),
                       'uniform_prop2': (uniform_crossoverGP_prop, 2),
                       'uniform_prop7': (uniform_crossoverGP_prop, 7),
                       'uniform_propk': (uniform_crossoverGP_prop, self.parents_num),
                       'uniform_rank2': (uniform_crossoverGP_rank, 2),
                       'uniform_rank7': (uniform_crossoverGP_rank, 7),
                       'uniform_rankk': (uniform_crossoverGP_rank, self.parents_num),
                       'uniform_tour3': (uniform_crossoverGP_tour, 3),
                       'uniform_tour7': (uniform_crossoverGP_tour, 7),
                       'uniform_tourk': (uniform_crossoverGP_tour, self.parents_num)}

        self.m_pool = {'weak_point': (point_mutation, 0.25, False),
                       'average_point': (point_mutation, 1, False),
                       'strong_point': (point_mutation, 4, False),
                       'custom_rate_point': (point_mutation, self.mutation_rate, True),
                       'weak_ephemeral': (ephemeral_mutation, 0.25, False),
                       'average_ephemeral': (ephemeral_mutation, 1, False),
                       'strong_ephemeral': (ephemeral_mutation, 4, False),
                       'custom_rate_ephemeral': (ephemeral_mutation, self.mutation_rate, True),
                       'weak_gauss': (ephemeral_gauss_mutation, 0.25, False),
                       'average_gauss': (ephemeral_gauss_mutation, 1, False),
                       'strong_gauss': (ephemeral_gauss_mutation, 4, False),
                       'custom_rate_gauss': (ephemeral_gauss_mutation, self.mutation_rate, True),
                       'weak_terminal': (terminal_mutation, 0.25, False),
                       'average_terminal': (terminal_mutation, 1, False),
                       'strong_terminal': (terminal_mutation, 4, False),
                       'custom_rate_terminal': (terminal_mutation, self.mutation_rate, True),
                       'weak_grow': (growing_mutation, 0.25, False),
                       'average_grow': (growing_mutation, 1, False),
                       'strong_grow': (growing_mutation, 4, False),
                       'custom_rate_grow': (growing_mutation, self.mutation_rate, True),
                       'weak_swap': (swap_mutation, 0.25, False),
                       'average_swap': (swap_mutation, 1, False),
                       'strong_swap': (swap_mutation, 4, False),
                       'custom_rate_swap': (swap_mutation, self.mutation_rate, True),
                       'weak_shrink': (shrink_mutation, 0.25, False),
                       'average_shrink': (shrink_mutation, 1, False),
                       'strong_shrink': (shrink_mutation, 4, False),
                       'custom_rate_shrink': (shrink_mutation, self.mutation_rate, True)}

    def set_strategy(self,
                     selection_oper = 'rank',
                     crossover_oper = 'standart',
                     mutation_oper = 'weak_grow',
                     tour_size_param = 2,
                     initial_population = None,
                     max_level = 16,
                     init_level = 5,
                     elitism_param = True,
                     parents_num_param = 7,
                     mutation_rate_param = 0.05):
        self.tour_size = tour_size_param
        self.initial_population = initial_population
        self.max_level = max_level
        self.init_level = init_level
        self.elitism = elitism_param
        self.parents_num = parents_num_param
        self.mutation_rate = mutation_rate_param

        self.update_pool()


        self.s_set = self.s_pool[selection_oper]
        self.c_set = self.c_pool[crossover_oper]
        self.m_set = self.m_pool[mutation_oper]

        return self

    def create_offspring(self, population_g, fitness_scale, fitness_rank, _):
        crossover_func, quantity = self.c_set
        selection_func, tour_size = self.s_set
        mutation_func, proba_down, not_scale = self.m_set

        indexes = selection_func(fitness_scale,
                                 fitness_rank,
                                 tour_size,
                                 quantity)

        parents = population_g[indexes]
        fitness_scale_p = fitness_scale[indexes]
        fitness_rank_p = fitness_rank[indexes]

        offspring_no_mutated = crossover_func(parents,
                                              fitness_scale_p,
                                              fitness_rank_p,
                                              self.max_level)

        if not_scale:
            proba = proba_down
        else:
            proba = proba_down/len(offspring_no_mutated)

        mutant = mutation_func(offspring_no_mutated,
                               self.uniset, proba, self.max_level)
        return mutant

    def fit(self):
        population_g = half_and_half(
            self.pop_size, self.uniset, self.init_level)
        population_ph = self.genotype_to_phenotype(population_g)
        fitness = self.evaluate(population_ph)
        fitness_scale = scale_data(fitness)
        fitness_rank = rank_data(fitness)

        self.thefittest = TheFittest().update(population_g,
                                              population_ph,
                                              fitness)
        lastbest = LastBest().update(self.thefittest.fitness)
        if self.keep_history is not None:
            self.stats = StatisticsGP(
                mode=self.keep_history).update(self.thefittest.genotype,
                                               fitness)
        for i in range(self.iters-1):
            self.show_progress(i)
            if self.termitation_check(lastbest.no_increase):
                break
            else:
                partial_create_offspring = partial(self.create_offspring,
                                                   population_g,
                                                   fitness_scale,
                                                   fitness_rank)
                map_ = map(partial_create_offspring, range(self.pop_size))
                population_g = np.array(list(map_), dtype=object)
                population_ph = self.genotype_to_phenotype(population_g)
                fitness = self.evaluate(population_ph)

                if self.elitism:
                    population_g[-1], population_ph[-1], fitness[-1] = self.thefittest.get()
                fitness_scale = scale_data(fitness)
                fitness_rank = rank_data(fitness)

                self.thefittest.update(population_g, population_ph, fitness)
                lastbest.update(self.thefittest.fitness)
                if self.keep_history is not None:
                    self.stats.update(self.thefittest.genotype,
                                      fitness)
        return self



import numpy as np
from ._base import TheFittest
from ._base import LastBest
from functools import partial
from ._differentialevolution import DifferentialEvolution
from ..tools.operators import binomial


class StatisticsjDE:
    def __init__(self, mode='quick'):
        self.mode = mode
        self.population_g = np.array([])
        self.fitness = np.array([])
        self.F = np.array([], dtype=float)
        self.CR = np.array([], dtype=float)

    def append_arr(self, arr_to, arr_from):
        shape_to = (-1, arr_from.shape[0], arr_from.shape[1])
        shape_from = (1, arr_from.shape[0], arr_from.shape[1])
        result = np.vstack([arr_to.reshape(shape_to),
                            arr_from.copy().reshape(shape_from)])
        return result

    def update(self,
               population_g_i: np.ndarray,
               fitness_i: np.ndarray,
               F_i, CR_i):
        if self.mode == 'quick':
            self.fitness = np.append(self.fitness, np.max(fitness_i))
        elif self.mode == 'full':
            self.fitness = np.append(self.fitness, np.max(fitness_i))
            self.population_g = self.append_arr(self.population_g,
                                                population_g_i)
            if not len(self.F):
                self.F = F_i.copy().reshape(1, -1)
                self.CR = CR_i.copy().reshape(1, -1)
            else:
                self.F = np.append(self.F, F_i.copy().reshape(1, -1), axis=0)
                self.CR = np.append(
                    self.CR, CR_i.copy().reshape(1, -1), axis=0)
        else:
            raise ValueError('the "mode" must be either "quick" or "full"')
        return self


class jDE(DifferentialEvolution):
    '''Brest, Janez & Greiner, Sao & Bošković, Borko & Mernik, Marjan & Zumer, Viljem. (2007).
    Self-Adapting Control Parameters in Differential Evolution: A Comparative Study on Numerical
    Benchmark Problems. Evolutionary Computation, IEEE Transactions on. 10. 646 - 657. 10.1109/TEVC.2006.872133. '''

    def __init__(self,
                 fitness_function,
                 genotype_to_phenotype,
                 iters,
                 pop_size,
                 left,
                 right,
                 optimal_value = None,
                 termination_error_value = 0.,
                 no_increase_num = None,
                 minimization = False,
                 show_progress_each = None,
                 keep_history = None):
        DifferentialEvolution.__init__(
            self,
            fitness_function=fitness_function,
            genotype_to_phenotype=genotype_to_phenotype,
            iters=iters,
            pop_size=pop_size,
            left=left,
            right=right,
            optimal_value=optimal_value,
            termination_error_value=termination_error_value,
            no_increase_num=no_increase_num,
            minimization=minimization,
            show_progress_each=show_progress_each,
            keep_history=keep_history)

        self.thefittest: TheFittest
        self.stats: StatisticsjDE
        self.F_left: float
        self.F_right: float
        self.t_f: float
        self.t_cr: float

    def set_strategy(self,
                     mutation_oper = 'rand_1',
                     F_left_param = 0.1,
                     F_right_param = 0.9,
                     t_f_param = 0.1,
                     t_cr_param = 0.1,
                     elitism_param = True,
                     initial_population = None):
        self.update_pool()
        self.m_function = self.m_pool[mutation_oper]
        self.F_left = F_left_param
        self.F_right = F_right_param
        self.t_f = t_cr_param
        self.t_cr = t_f_param
        self.elitism = elitism_param
        self.initial_population = initial_population
        return self

    def mutation_and_crossover(self, popuation_g, individ_g, F_i, CR_i):
        mutant = self.m_function(individ_g, popuation_g, F_i)

        mutant_cr_g = binomial(individ_g, mutant, CR_i)
        mutant_cr_g = self.bounds_control(mutant_cr_g)
        return mutant_cr_g

    def evaluate_and_selection(self, mutant_cr_g, population_g, population_ph, fitness):
        offspring_g = population_g.copy()
        offspring_ph = population_ph.copy()
        offspring_fit = fitness.copy()

        mutant_cr_ph = self.genotype_to_phenotype(mutant_cr_g)
        mutant_cr_fit = self.evaluate(mutant_cr_ph)
        mask = mutant_cr_fit >= fitness
        offspring_g[mask] = mutant_cr_g[mask]
        offspring_ph[mask] = mutant_cr_ph[mask]
        offspring_fit[mask] = mutant_cr_fit[mask]
        return offspring_g, offspring_ph, offspring_fit, mask

    def bounds_control(self, individ_g):
        low_mask = individ_g < self.left
        high_mask = individ_g > self.right

        individ_g[low_mask] = self.left[low_mask]
        individ_g[high_mask] = self.right[high_mask]
        return individ_g

    def regenerate_F(self, F_i):
        mask = np.random.random(size=len(F_i)) < self.t_f
        F_i[mask] = self.F_left + \
            np.random.random(size=np.sum(mask))*self.F_right
        return F_i

    def regenerate_CR(self, CR_i):
        mask = np.random.random(size=len(CR_i)) < self.t_cr
        CR_i[mask] = np.random.random(size=np.sum(mask))
        return CR_i

    def fit(self):
        population_g = self.generate_init_pop()
        population_ph = self.genotype_to_phenotype(population_g)
        fitness = self.evaluate(population_ph)

        argsort = np.argsort(fitness)
        population_g = population_g[argsort]
        population_ph = population_ph[argsort]
        fitness = fitness[argsort]

        F_i = np.full(self.pop_size, 0.5)
        CR_i = np.full(self.pop_size, 0.9)

        self.thefittest = TheFittest().update(population_g,
                                              population_ph,
                                              fitness)
        lastbest = LastBest().update(self.thefittest.fitness)
        if self.keep_history is not None:
            self.stats = StatisticsjDE(
                mode=self.keep_history).update(population_g,
                                               fitness,
                                               F_i,
                                               CR_i)

        for i in range(self.iters-1):
            self.show_progress(i)
            if self.termitation_check(lastbest.no_increase):
                break
            else:

                F_i_new = self.regenerate_F(F_i.copy())
                CR_i_new = self.regenerate_CR(CR_i.copy())

                partial_mut_and_cross = partial(self.mutation_and_crossover,
                                                population_g)
                mutant_cr_g = np.array(list(map(partial_mut_and_cross,
                                                population_g,
                                                F_i_new, CR_i_new)))

                stack = self.evaluate_and_selection(mutant_cr_g,
                                                    population_g,
                                                    population_ph,
                                                    fitness)
                population_g = stack[0]
                population_ph = stack[1]
                fitness = stack[2]

                succeses = stack[3]
                F_i[succeses] = F_i_new[succeses]
                CR_i[succeses] = CR_i_new[succeses]

                if self.elitism:
                    population_g[-1], population_ph[-1], fitness[-1] = self.thefittest.get()
                argsort = np.argsort(fitness)
                population_g = population_g[argsort]
                population_ph = population_ph[argsort]
                fitness = fitness[argsort]

                self.thefittest.update(population_g, population_ph, fitness)
                lastbest.update(self.thefittest.fitness)
                if self.keep_history is not None:
                    self.stats.update(population_g,
                                      fitness,
                                      F_i,
                                      CR_i)
        return self



import numpy as np
from typing import Callable
from ._base import TheFittest
from ._base import EvolutionaryAlgorithm
from ._base import LastBest
from functools import partial
from ..tools.operators import binomial
from ..tools.operators import best_1
from ..tools.operators import best_2
from ..tools.operators import rand_to_best1
from ..tools.operators import current_to_best_1
from ..tools.operators import rand_1
from ..tools.operators import current_to_pbest_1
from ..tools.operators import rand_2
from ..tools.generators import float_population


class Statistics:
    def __init__(self, mode='quick'):
        self.mode = mode
        self.population_g = np.array([])
        self.fitness = np.array([])

    def append_arr(self, arr_to, arr_from):
        shape_to = (-1, arr_from.shape[0], arr_from.shape[1])
        shape_from = (1, arr_from.shape[0], arr_from.shape[1])
        result = np.vstack([arr_to.reshape(shape_to),
                            arr_from.copy().reshape(shape_from)])
        return result

    def update(self,
               population_g_i,
               fitness_i):
        if self.mode == 'quick':
            self.fitness = np.append(self.fitness, np.max(fitness_i))
        elif self.mode == 'full':
            self.fitness = np.append(self.fitness, np.max(fitness_i))
            self.population_g = self.append_arr(self.population_g,
                                                population_g_i)
        else:
            raise ValueError('the "mode" must be either "quick" or "full"')
        return self


class DifferentialEvolution(EvolutionaryAlgorithm):
    '''Storn, Rainer & Price, Kenneth. (1995). Differential Evolution: A Simple and Efficient
    Adaptive Scheme for Global Optimization Over Continuous Spaces. Journal of Global Optimization. 23'''

    def __init__(self,
                 fitness_function,
                 genotype_to_phenotype,
                 iters,
                 pop_size,
                 left,
                 right,
                 optimal_value = None,
                 termination_error_value = 0.,
                 no_increase_num = None,
                 minimization = False,
                 show_progress_each = None,
                 keep_history = None):
        EvolutionaryAlgorithm.__init__(
            self,
            fitness_function=fitness_function,
            genotype_to_phenotype=genotype_to_phenotype,
            iters=iters,
            pop_size=pop_size,
            optimal_value=optimal_value,
            termination_error_value=termination_error_value,
            no_increase_num=no_increase_num,
            minimization=minimization,
            show_progress_each=show_progress_each,
            keep_history=keep_history)

        self.left = left
        self.right = right
        self.thefittest: TheFittest
        self.stats: Statistics
        self.initial_population: np.ndarray
        self.m_function: Callable
        self.F: float
        self.CR: float
        self.elitism: bool

        self.set_strategy()

    def generate_init_pop(self):
        if self.initial_population is None:
            population_g = float_population(
                self.pop_size, self.left, self.right)
        else:
            population_g = self.initial_population
        return population_g

    def update_pool(self):
        self.m_pool = {'best_1': best_1,
                       'rand_1': rand_1,
                       'current_to_best_1': current_to_best_1,
                       'current_to_pbest_1': current_to_pbest_1,
                       'rand_to_best1': rand_to_best1,
                       'best_2': best_2,
                       'rand_2': rand_2}

    def set_strategy(self,
                     mutation_oper = 'rand_1',
                     F_param = 0.5,
                     CR_param = 0.5,
                     elitism_param = True,
                     initial_population = None):
        
        self.update_pool()
        self.m_function = self.m_pool[mutation_oper]
        self.F = F_param
        self.CR = CR_param
        self.elitism = elitism_param
        self.initial_population = initial_population
        return self

    def mutation_and_crossover(self, popuation_g, individ_g, F_i, CR_i):
        mutant = self.m_function(individ_g, popuation_g, F_i)

        mutant_cr_g = binomial(individ_g, mutant, CR_i)
        mutant_cr_g = self.bounds_control(mutant_cr_g)
        return mutant_cr_g

    def evaluate_and_selection(self, mutant_cr_g, population_g, population_ph, fitness):
        offspring_g = population_g.copy()
        offspring_ph = population_ph.copy()
        offspring_fit = fitness.copy()

        mutant_cr_ph = self.genotype_to_phenotype(mutant_cr_g)
        mutant_cr_fit = self.evaluate(mutant_cr_ph)
        mask = mutant_cr_fit >= fitness
        offspring_g[mask] = mutant_cr_g[mask]
        offspring_ph[mask] = mutant_cr_ph[mask]
        offspring_fit[mask] = mutant_cr_fit[mask]
        return offspring_g, offspring_ph, offspring_fit

    def bounds_control(self, individ_g):
        individ_g = individ_g.copy()
        low_mask = individ_g < self.left
        high_mask = individ_g > self.right

        individ_g[low_mask] = self.left[low_mask]
        individ_g[high_mask] = self.right[high_mask]
        return individ_g

    def fit(self):
        population_g = self.generate_init_pop()
        population_ph = self.genotype_to_phenotype(population_g)
        fitness = self.evaluate(population_ph)

        argsort = np.argsort(fitness)
        population_g = population_g[argsort]
        population_ph = population_ph[argsort]
        fitness = fitness[argsort]

        self.thefittest = TheFittest().update(population_g,
                                              population_ph,
                                              fitness)
        lastbest = LastBest().update(self.thefittest.fitness)
        if self.keep_history is not None:
            self.stats = Statistics(
                mode=self.keep_history).update(population_g,
                                               fitness)

        for i in range(self.iters-1):
            self.show_progress(i)
            if self.termitation_check(lastbest.no_increase):
                break
            else:
                F_i = np.full(self.pop_size, self.F)
                CR_i = np.full(self.pop_size, self.CR)

                partial_mut_and_cross = partial(self.mutation_and_crossover,
                                                population_g)
                mutant_cr_g = np.array(list(map(partial_mut_and_cross,
                                                population_g,
                                                F_i, CR_i)))

                stack = self.evaluate_and_selection(mutant_cr_g,
                                                    population_g,
                                                    population_ph,
                                                    fitness)
                population_g, population_ph, fitness = stack

                if self.elitism:
                    population_g[-1], population_ph[-1], fitness[-1] = self.thefittest.get()
                argsort = np.argsort(fitness)
                population_g = population_g[argsort]
                population_ph = population_ph[argsort]
                fitness = fitness[argsort]

                self.thefittest.update(population_g, population_ph, fitness)
                lastbest.update(self.thefittest.fitness)
                if self.keep_history is not None:
                    self.stats.update(population_g,
                                      fitness)
        return self



import numpy as np


def root_mean_square_error(y_true, y_predict):
    return np.sqrt(np.mean((y_true - y_predict)**2))

def coefficient_determination(y_true, y_predict):
    mean = np.mean(y_true)
    residual_sum = np.sum((y_true - y_predict)**2)
    total_sum = np.sum((y_true - mean)**2)
    return 1 - residual_sum/total_sum


def common_region(trees):
    terminate = False
    indexes = []
    common_indexes = []
    border_indexes = []
    for tree in trees:
        indexes.append(list(range(len(tree.nodes))))
        common_indexes.append([])
        border_indexes.append([])

    while not terminate:
        inner_break = False
        iters = np.min(list(map(len, indexes)))

        for i in range(iters):
            first_n_args = trees[0].nodes[indexes[0][i]].n_args
            common_indexes[0].append(indexes[0][i])
            for j in range(1, len(indexes)):
                common_indexes[j].append(indexes[j][i])
                if first_n_args != trees[j].nodes[indexes[j][i]].n_args:
                    inner_break = True

            if inner_break:
                for j in range(0, len(indexes)):
                    border_indexes[j].append(indexes[j][i])
                break

        for j in range(len(indexes)):
            _, right = trees[j].subtree(common_indexes[j][-1])
            delete_to = indexes[j].index(right-1) + 1
            indexes[j] = indexes[j][delete_to:]

            if len(indexes[j]) < 1:
                terminate = True
                break

    return common_indexes, border_indexes


def donothing(x):
    return x


def numpy_group_by(group, by):

    argsort = np.argsort(by)
    group = group[argsort]
    by = by[argsort]

    keys, cut_index = np.unique(by, return_index=True)
    groups = np.split(group, cut_index)[1:]
    return keys, groups


def lehmer_mean(x, power=2):
    x_up = np.power(x, power)
    x_down = np.power(x, power-1)
    return np.sum(x_up)/np.sum(x_down)


def rank_data(arr):
    arr = arr.copy()
    raw_ranks = np.zeros(shape=(arr.shape[0]))
    argsort = np.argsort(arr)

    s = (arr[:, np.newaxis] == arr).astype(int)
    raw_ranks[argsort] = np.arange(arr.shape[0]) + 1
    ranks = np.sum(raw_ranks*s, axis=1)/s.sum(axis=0)
    return ranks


def protect_norm(x):
    sum_ = x.sum()
    if sum_ > 0:
        return x/sum_
    else:
        len_ = len(x)
        return np.full(len_, 1/len_)


def scale_data(arr):
    arr = arr.copy()
    max_ = arr.max()
    min_ = arr.min()
    if max_ == min_:
        arr_n = np.ones_like(arr)
    else:
        arr_n = (arr - min_)/(max_ - min_)
    return arr_n


def numpy_bit_to_int(bit_array, powers=None):
    if powers is None:
        powers = 2**np.arange(bit_array.shape[1], dtype=np.int64)
    arange_ = powers[:bit_array.shape[1]][::-1]
    int_array = np.sum(bit_array*arange_, axis=1)
    return int_array


def numpy_int_to_bit(int_array):
    result = []
    bit = int_array % 2
    remains = int_array//2
    result.append(bit)
    while np.sum(remains) > 0:
        bit = remains % 2
        remains = remains//2
        result.append(bit)

    return np.array(result)[::-1].T


def numpy_gray_to_bit(gray_array):
    bit_array = np.logical_xor.accumulate(
        gray_array, axis=-1).astype(np.byte)
    return bit_array


def numpy_bit_to_gray(bit_array):
    cut_gray = np.logical_xor(bit_array[:, :-1],
                              bit_array[:, 1:])
    gray_array = np.hstack(
        [bit_array[:, 0].reshape(-1, 1), cut_gray])
    return gray_array


class SamplingGrid:

    def __init__(self, fit_by = 'h') -> None:
        self.fit_by = fit_by
        self.left: np.ndarray
        self.right: np.ndarray
        self.parts: np.ndarray
        self.h: np.ndarray
        self.power_arange: np.ndarray

    def culc_h_from_parts(self, left, right, parts):
        return (right - left)/(2.0**parts - 1)

    def culc_parts_from_h(self, left, right, h):
        return np.ceil(np.log2((right - left)/h + 1)).astype(int)

    def decode(self, bit_array_i):
        int_convert = numpy_bit_to_int(bit_array_i, self.power_arange)
        return int_convert

    def fit(self, left, right,
            arg):
        self.left = left
        self.right = right

        assert self.fit_by in [
            'h', 'parts'], f"incorrect option {self.fit_by} for fit_by. The available ones are 'h' and 'parts'"
        if self.fit_by == 'h':
            min_h = arg
            self.parts = self.culc_parts_from_h(left, right, min_h)
            self.h = self.culc_h_from_parts(left, right, self.parts)
        else:
            self.parts = arg
            self.h = self.culc_h_from_parts(left, right, self.parts)

        self.power_arange = 2**np.arange(self.parts.max(), dtype=np.int64)
        return self

    def transform(self, population):
        splits = np.add.accumulate(self.parts)
        p_parts = np.split(population, splits[:-1], axis=1)

        int_array = np.array(list(map(self.decode, p_parts))).T
        float_array = self.left[np.newaxis, :] +\
            self.h[np.newaxis, :]*int_array

        return float_array

    def float_to_bit(self, float_array, left, h):
        grid_number = (float_array - left)/h
        int_array = np.rint(grid_number)
        return numpy_int_to_bit(int_array)

    def inverse_transform(self, population):
        map_ = map(self.float_to_bit, population.T, self.left, self.h)
        return np.hstack(list(map_))


class GrayCode(SamplingGrid):

    def __init__(self, fit_by: str = 'h') -> None:
        SamplingGrid.__init__(self, fit_by)

    def decode(self, gray_array_i):
        bit_array_i = numpy_gray_to_bit(gray_array_i)
        int_convert = numpy_bit_to_int(bit_array_i, self.power_arange)
        return int_convert

    def float_to_bit(self, float_array, left, h):
        grid_number = (float_array - left)/h
        int_array = np.rint(grid_number)
        bit_array = numpy_int_to_bit(int_array)
        return numpy_bit_to_gray(bit_array)



import numpy as np
import os


path = os.path.dirname(__file__) + '/shifts_data/'

fbias_data = np.loadtxt(path+'fbias_data.txt')
# 1
sphere_func_data = np.loadtxt(path+'sphere_func_data.txt')
# 2
schwefel_102_data = np.loadtxt(path+'schwefel_102_data.txt')
# 3
high_cond_elliptic_rot_data = np.loadtxt(path
                                         + 'high_cond_elliptic_rot_data.txt')[:50]
elliptic_M_D2 = np.loadtxt(path+'elliptic_M_D2.txt')
elliptic_M_D10 = np.loadtxt(path+'elliptic_M_D10.txt')
elliptic_M_D30 = np.loadtxt(path+'elliptic_M_D30.txt')
elliptic_M_D50 = np.loadtxt(path+'elliptic_M_D50.txt')
# 5
schwefel_206_data = np.loadtxt(path+'schwefel_206_data.txt')
o_206 = schwefel_206_data[0]
A_206 = schwefel_206_data[1:]
# 6
rosenbrock_func_data = np.loadtxt(path+'rosenbrock_func_data.txt')
# 7
griewank_func_data = np.loadtxt(path+'griewank_func_data.txt')[:50]
griewank_M_D2 = np.loadtxt(path+'griewank_M_D2.txt')
griewank_M_D10 = np.loadtxt(path+'griewank_M_D10.txt')
griewank_M_D30 = np.loadtxt(path+'griewank_M_D30.txt')
griewank_M_D50 = np.loadtxt(path+'griewank_M_D50.txt')
# 8
ackley_func_data = np.loadtxt(path+'ackley_func_data.txt')[:50]
ackley_M_D2 = np.loadtxt(path+'ackley_M_D2.txt')
ackley_M_D10 = np.loadtxt(path+'ackley_M_D10.txt')
ackley_M_D30 = np.loadtxt(path+'ackley_M_D30.txt')
ackley_M_D50 = np.loadtxt(path+'ackley_M_D50.txt')
# 9-10
rastrigin_func_data = np.loadtxt(path+'rastrigin_func_data.txt')[:50]
rastrigin_M_D2 = np.loadtxt(path+'rastrigin_M_D2.txt')
rastrigin_M_D10 = np.loadtxt(path+'rastrigin_M_D10.txt')
rastrigin_M_D30 = np.loadtxt(path+'rastrigin_M_D30.txt')
rastrigin_M_D50 = np.loadtxt(path+'rastrigin_M_D50.txt')
# 11
weierstrass_M_D2 = np.loadtxt(path+'weierstrass_M_D2.txt')
weierstrass_M_D10 = np.loadtxt(path+'weierstrass_M_D10.txt')
weierstrass_M_D30 = np.loadtxt(path+'weierstrass_M_D30.txt')
weierstrass_M_D50 = np.loadtxt(path+'weierstrass_M_D50.txt')
weierstrass_data = np.loadtxt(path+'weierstrass_data.txt')[:50]
# 12
schwefel_213_data = np.loadtxt(path+'schwefel_213_data.txt')
# 13
EF8F2_func_data = np.loadtxt(path+'EF8F2_func_data.txt')
a_213 = schwefel_213_data[:100]
b_213 = schwefel_213_data[100:200]
alpha_213 = schwefel_213_data[-1]
# 14
E_ScafferF6_func_data = np.loadtxt(path+'E_ScafferF6_func_data.txt')[:50]
E_ScafferF6_M_D2 = np.loadtxt(path+'E_ScafferF6_M_D2.txt')
E_ScafferF6_M_D10 = np.loadtxt(path+'E_ScafferF6_M_D10.txt')
E_ScafferF6_M_D30 = np.loadtxt(path+'E_ScafferF6_M_D30.txt')
E_ScafferF6_M_D50 = np.loadtxt(path+'E_ScafferF6_M_D50.txt')
# 15
hybrid_func1_data = np.loadtxt(path+'hybrid_func1_data.txt')
# 16
hybrid_func1_M_D2 = np.loadtxt(path+'hybrid_func1_M_D2.txt')
hybrid_func1_M_D10 = np.loadtxt(path+'hybrid_func1_M_D10.txt')
hybrid_func1_M_D30 = np.loadtxt(path+'hybrid_func1_M_D30.txt')
hybrid_func1_M_D50 = np.loadtxt(path+'hybrid_func1_M_D50.txt')
# 18
hybrid_func2_data = np.loadtxt(path+'hybrid_func2_data.txt')
hybrid_func2_M_D2 = np.loadtxt(path+'hybrid_func2_M_D2.txt')
hybrid_func2_M_D10 = np.loadtxt(path+'hybrid_func2_M_D10.txt')
hybrid_func2_M_D30 = np.loadtxt(path+'hybrid_func2_M_D30.txt')
hybrid_func2_M_D50 = np.loadtxt(path+'hybrid_func2_M_D50.txt')
# 21
hybrid_func3_data = np.loadtxt(path+'hybrid_func3_data.txt')
hybrid_func3_M_D2 = np.loadtxt(path+'hybrid_func3_M_D2.txt')
hybrid_func3_M_D10 = np.loadtxt(path+'hybrid_func3_M_D10.txt')
hybrid_func3_M_D30 = np.loadtxt(path+'hybrid_func3_M_D30.txt')
hybrid_func3_M_D50 = np.loadtxt(path+'hybrid_func3_M_D50.txt')
# 22
hybrid_func3_HM_D2 = np.loadtxt(path+'hybrid_func3_HM_D2.txt')
hybrid_func3_HM_D10 = np.loadtxt(path+'hybrid_func3_HM_D10.txt')
hybrid_func3_HM_D30 = np.loadtxt(path+'hybrid_func3_HM_D30.txt')
hybrid_func3_HM_D50 = np.loadtxt(path+'hybrid_func3_HM_D50.txt')
# 24
hybrid_func4_data = np.loadtxt(path+'hybrid_func4_data.txt')
hybrid_func4_M_D2 = np.loadtxt(path+'hybrid_func4_M_D2.txt')
hybrid_func4_M_D10 = np.loadtxt(path+'hybrid_func4_M_D10.txt')
hybrid_func4_M_D30 = np.loadtxt(path+'hybrid_func4_M_D30.txt')
hybrid_func4_M_D50 = np.loadtxt(path+'hybrid_func4_M_D50.txt')


class TestFunction:
    def __call__(self, x):
        return self.f(x)

    def build_grid(self, x, y):
        x1, y1 = np.meshgrid(x, y)
        xy = np.concatenate(
            [x1[:, :, np.newaxis], y1[:, :, np.newaxis]], axis=2)
        z = np.zeros(shape=xy.shape[:-1])
        for i, x_i in enumerate(xy):
            z[i] = self(x_i)
        return z


class TestShiftedFunction:
    def __init__(self, fbias, x_shift):
        self.fbias = fbias
        self.x_shift = x_shift

    def shift(self, x):
        shape = x.shape
        axis = [1]*(len(shape)-1) + [-1]
        return x - self.x_shift[:shape[-1]].reshape(axis)

    def __call__(self, x):
        z = self.shift(x)
        return self.f(z) + self.fbias

    def build_grid(self, x, y):
        x1, y1 = np.meshgrid(x, y)
        xy = np.concatenate(
            [x1[:, :, np.newaxis], y1[:, :, np.newaxis]], axis=2)
        z = np.zeros(shape=xy.shape[:-1])
        for i, x_i in enumerate(xy):
            z[i] = self(x_i)
        return z


class TestShiftedRotatedFunction:
    def __init__(self, fbias, x_shift, rotate_M_D2,
                 rotate_M_D10, rotate_M_D30, rotate_M_D50):
        self.fbias = fbias
        self.x_shift = x_shift
        self.rotate_M_D2 = rotate_M_D2
        self.rotate_M_D10 = rotate_M_D10
        self.rotate_M_D30 = rotate_M_D30
        self.rotate_M_D50 = rotate_M_D50

    def shift(self, x):
        shape = x.shape
        axis = [1]*(len(shape)-1) + [-1]
        return x - self.x_shift[:shape[-1]].reshape(axis)

    def rotate(self, x):
        if x.shape[1] == 2:
            z = x@self.rotate_M_D2
        elif x.shape[1] == 10:
            z = x@self.rotate_M_D10
        elif x.shape[1] == 30:
            z = x@self.rotate_M_D30
        elif x.shape[1] == 50:
            z = x@self.rotate_M_D50
        return z

    def __call__(self, x):
        z = self.shift(x)
        z_rotated = self.rotate(z)
        return self.f(z_rotated) + self.fbias

    def build_grid(self, x, y):
        x1, y1 = np.meshgrid(x, y)
        xy = np.concatenate(
            [x1[:, :, np.newaxis], y1[:, :, np.newaxis]], axis=2)
        z = np.zeros(shape=xy.shape[:-1])
        for i, x_i in enumerate(xy):
            z[i] = self(x_i)
        return z


class SampleHybridCompositionFunction():
    def __init__(self, basic_functions, sigmas, lambdas,
                 fbias,
                 x_shift,
                 M_D2, M_D10, M_D30, M_D50):
        self.basic_functions = basic_functions
        self.sigmas = sigmas
        self.lambdas = lambdas
        self.M_D2 = np.split(M_D2, range(
            M_D2.shape[1], M_D2.shape[0], M_D2.shape[1]))
        self.M_D10 = np.split(M_D10, range(
            M_D10.shape[1], M_D10.shape[0], M_D10.shape[1]))
        self.M_D30 = np.split(M_D30, range(
            M_D30.shape[1], M_D30.shape[0], M_D30.shape[1]))
        self.M_D50 = np.split(M_D50, range(
            M_D50.shape[1], M_D50.shape[0], M_D50.shape[1]))
        self.fbias = fbias
        self.x_shift = x_shift

        self.bias = np.array(
            [0, 100, 200, 300, 400, 500, 600, 700, 800, 900], dtype=np.float64)
        self.C = 2000

    def shift(self, x, i):
        shape = x.shape
        axis = [1]*(len(shape)-1) + [-1]
        return x - self.x_shift[i][:shape[-1]].reshape(axis)

    def rotate(self, x, M_D2, M_D10, M_D30, M_D50):
        if x.shape[1] == 2:
            z = x@M_D2
        elif x.shape[1] == 10:
            z = x@M_D10
        elif x.shape[1] == 30:
            z = x@M_D30
        elif x.shape[1] == 50:
            z = x@M_D50
        return z

    def calc_w(self, x, shift_i, sigma_i, i):
        D = x.shape[1]
        shift_x = self.shift(x, i)

        up = np.sum(shift_x**2, axis=-1)
        down = 2*D*(sigma_i*sigma_i)

        w_i = np.exp(-(up/down))

        return w_i, shift_x

    def procces_i_function(self, x, func_i, sigma_i, lambda_i, shift_i,
                           M_D2i, M_D10i, M_D30i, M_D50i, i):
        y = np.full(shape=x.shape, fill_value=5)
        w_i, shift_x = self.calc_w(x, shift_i, sigma_i, i)

        arg1 = self.rotate(shift_x/lambda_i,  M_D2i, M_D10i, M_D30i, M_D50i)
        fit_i = func_i()(arg1)

        arg2 = self.rotate(y/lambda_i,  M_D2i, M_D10i, M_D30i, M_D50i)
        f_max_i = func_i()(arg2)

        fit_i = self.C*fit_i/f_max_i

        return w_i, fit_i

    def __call__(self, x):
        result = tuple(self.procces_i_function(x, func_i, sigma_i, lambda_i, shift_i,
                                               M_D2i, M_D10i, M_D30i, M_D50i, i)
                       for func_i, sigma_i, lambda_i, shift_i,
                       M_D2i, M_D10i, M_D30i, M_D50i, i in zip(self.basic_functions, self.sigmas, self.lambdas, self.x_shift,
                                                               self.M_D2, self.M_D10, self.M_D30, self.M_D50, range(len(self.M_D50))))
        w_i, fit_i = list(zip(*result))
        w_i = np.array(w_i, np.float64).T
        fit_i = np.array(fit_i, np.float64).T
        max_w = np.max(w_i, axis=1)

        for i, w_i_j in enumerate(w_i.T):
            cond = w_i_j != max_w
            w_i[:, i][cond] = w_i[:, i][cond]*(1.0 - max_w[cond]**10)

        w_i = w_i/np.sum(w_i, axis=-1)[:, np.newaxis]

        return np.sum(w_i*(fit_i + self.bias), axis=-1) + self.fbias

    def test(self):
        y = self(self.x_shift.reshape(1, -1))
        return y - self.fbias < self.fixed_accuracy

    def build_grid(self, x, y):
        x1, y1 = np.meshgrid(x, y)
        xy = np.concatenate(
            [x1[:, :, np.newaxis], y1[:, :, np.newaxis]], axis=2)
        z = np.zeros(shape=xy.shape[:-1])
        for i, x_i in enumerate(xy):
            z[i] = self(x_i)
        return z


class OneMax(TestFunction):
    def f(self, x):
        return np.sum(x, axis=1)


class Sphere(TestFunction):
    def f(self, x):
        return np.sum(x**2, axis=-1)


class Schwefe1_2(TestFunction):
    def f(self, x):
        return np.sum(np.add.accumulate(x, axis=-1)**2, axis=-1)


class HighConditionedElliptic(TestFunction):
    def f(self, x):
        i = np.arange(1, x.shape[1]+1)
        demension = x.shape[1]
        return np.sum((1e6**((i - 1)/(demension - 1)))*x**2, axis=-1)


class Rosenbrock(TestFunction):
    def f(self, x):
        value = 100*((x.T[:-1]**2 - x.T[1:])**2) + (x.T[:-1] - 1)**2
        return np.sum(value.T, axis=-1)


class Rastrigin(TestFunction):
    def f(self, x):
        return np.sum(x**2 - 10*np.cos(2*np.pi*x) + 10, axis=-1)


class Griewank(TestFunction):
    def f(self, x):
        sqrt_i = np.sqrt(np.arange(1, x.shape[1]+1))
        sum_ = np.sum((x**2)/4000, axis=-1)
        prod_ = np.prod(np.cos(x/sqrt_i), axis=-1)
        return sum_ - prod_ + 1


class Ackley(TestFunction):
    def f(self, x):
        a = 20
        c = 2*np.pi
        b = 0.2
        D = x.shape[1]
        left = -a*np.exp(-b*np.sqrt(np.sum(x**2, axis=1)/D))
        right = -np.exp((1/D)*np.sum(np.cos(c*x), axis=1))
        return left + right + a + np.exp(1)


class Weierstrass(TestFunction):
    def f(self, x):
        a = 0.5
        b = 3
        k_max = np.arange(20, dtype=np.int64)
        D = x.shape[1]

        two_pi_power_b = 2*np.pi*(b**k_max)
        a_power_k = a**k_max
        arg1 = two_pi_power_b*(x[:, :, np.newaxis]+0.5)
        arg2 = two_pi_power_b*0.5
        left = a_power_k*np.cos(arg1)
        right = a_power_k*np.cos(arg2)

        left_sum = np.sum(np.sum(left, axis=-1), axis=-1)
        right_sum = np.sum(right, axis=-1)
        return left_sum - D*right_sum


class F8F2(TestFunction):
    def __init__(self):
        self.rosenbrock_f = Rosenbrock()
        self.griewank_f = Griewank()

    def f(self, x):
        indexes = np.kron(
            np.arange(1, x.shape[1]-1, dtype=np.int64), np.array([1, 1]))
        indexes = np.insert(indexes, [0], [0])
        indexes = np.append(indexes, np.array([x.shape[1]-1, x.shape[1]-1, 0]))

        x_indexes = x[:, indexes]

        vertical_X = x_indexes.reshape(-1, 2)
        F2 = self.rosenbrock_f(vertical_X)
        F8 = self.griewank_f(F2.reshape(-1, 1))
        horizontal_X = F8.reshape(x.shape[0], -1)
        return np.sum(horizontal_X, axis=-1)


class ExpandedScaffers_F6(TestFunction):
    def Scaffes_F6(self, x):
        sum_of_power = x[:, 0]**2 + x[:, 1]**2
        up = np.sin(np.sqrt(sum_of_power))**2 - 0.5
        down = (1 + 0.001*(sum_of_power))**2
        return 0.5 + up/down

    def f(self, x):
        indexes = np.kron(
            np.arange(1, x.shape[1]-1, dtype=np.int64), np.array([1, 1]))
        indexes = np.insert(indexes, [0], [0])
        indexes = np.append(indexes, np.array([x.shape[1]-1, x.shape[1]-1, 0]))
        x_indexes = x[:, indexes]

        vertical_X = x_indexes.reshape(-1, 2)
        F = self.Scaffes_F6(vertical_X)
        horizontal_X = F.reshape(x.shape[0], -1)

        return np.sum(horizontal_X, axis=-1)


class NonContinuosRastrigin(Rastrigin):
    def __call__(self, x):
        cond = np.abs(x) >= 0.5
        x[cond] = (np.round(2*x)/2)[cond]
        return super().__call__(x)


class NonContinuosExpandedScaffers_F6(ExpandedScaffers_F6):
    def __call__(self, x):
        cond = np.abs(x) >= 0.5
        x[cond] = (np.round(2*x)/2)[cond]
        return super().__call__(x)


class SphereWithNoise(Sphere):
    def __call__(self, x):
        y = super().__call__(x)
        noise = 1 + 0.1*np.abs(np.random.normal(size=y.shape))
        return y*noise


# CEC05 #1
class ShiftedSphere(TestShiftedFunction, Sphere):
    def __init__(self):
        TestShiftedFunction.__init__(self,
                                     fbias=fbias_data[0],
                                     x_shift=sphere_func_data)


# CEC05 #2
class ShiftedSchwefe1_2(TestShiftedFunction, Schwefe1_2):
    def __init__(self):
        TestShiftedFunction.__init__(self,
                                     fbias=fbias_data[1],
                                     x_shift=schwefel_102_data)


# CEC05 #3
class ShiftedRotatedHighConditionedElliptic(TestShiftedRotatedFunction,
                                            HighConditionedElliptic):
    def __init__(self):
        TestShiftedRotatedFunction.__init__(
            self,
            fbias=fbias_data[2],
            x_shift=high_cond_elliptic_rot_data,
            rotate_M_D2=elliptic_M_D2,
            rotate_M_D10=elliptic_M_D10,
            rotate_M_D30=elliptic_M_D30,
            rotate_M_D50=elliptic_M_D50)


# CEC05 #4
class ShiftedSchwefe1_2WithNoise(TestShiftedFunction, Schwefe1_2):
    def __init__(self):
        TestShiftedFunction.__init__(self,
                                     fbias=fbias_data[3],
                                     x_shift=schwefel_102_data)

    def __call__(self, x):
        z = self.shift(x)
        y = self.f(z)
        noise = 1 + 0.4*np.abs(np.random.normal(size=y.shape))
        return y*noise + self.fbias


# CEC05 #5
class Schwefel2_6(TestShiftedFunction):
    def __init__(self):
        TestShiftedFunction.__init__(self,
                                     fbias=fbias_data[4],
                                     x_shift=o_206)

    def __call__(self, x):
        D = x.shape[1]
        self.x_shift[np.floor(3*D/4).astype(np.int64)-1:D] = 100
        self.x_shift[:np.ceil(D/4).astype(np.int64)] = -100
        return self.f(x) + self.fbias

    def f(self, x):
        o = self.x_shift[:x.shape[1]]
        A = A_206[:x.shape[1], :x.shape[1]]
        Ax = A@x.T
        B = A@o
        fx = np.abs(Ax - B[:, np.newaxis])
        return np.max(fx, axis=0)


# CEC05 #6
class ShiftedRosenbrock(TestShiftedFunction, Rosenbrock):
    def __init__(self):
        TestShiftedFunction.__init__(self,
                                     fbias=fbias_data[5],
                                     x_shift=rosenbrock_func_data)

    def shift(self, x):
        shape = x.shape
        axis = [1]*(len(shape)-1) + [-1]
        return x - self.x_shift[:shape[-1]].reshape(axis) + 1


# CEC05 #7
class ShiftedRotatedGriewank(TestShiftedRotatedFunction,
                             Griewank):
    def __init__(self):
        TestShiftedRotatedFunction.__init__(
            self,
            fbias=fbias_data[6],
            x_shift=griewank_func_data,
            rotate_M_D2=griewank_M_D2,
            rotate_M_D10=griewank_M_D10,
            rotate_M_D30=griewank_M_D30,
            rotate_M_D50=griewank_M_D50)


# CEC05 #8
class ShiftedRotatedAckley(TestShiftedRotatedFunction,
                           Ackley):
    def __init__(self):
        TestShiftedRotatedFunction.__init__(
            self,
            fbias=fbias_data[7],
            x_shift=ackley_func_data,
            rotate_M_D2=ackley_M_D2,
            rotate_M_D10=ackley_M_D10,
            rotate_M_D30=ackley_M_D30,
            rotate_M_D50=ackley_M_D50)

    def __call__(self, x):
        D = x.shape[1]
        self.x_shift[:D:2] = -32.0
        return super().__call__(x)


# CEC05 #9
class ShiftedRastrigin(TestShiftedFunction, Rastrigin):
    def __init__(self):
        TestShiftedFunction.__init__(self,
                                     fbias=fbias_data[8],
                                     x_shift=rastrigin_func_data)


# CEC05 #10
class ShiftedRotatedRastrigin(TestShiftedRotatedFunction,
                              Rastrigin):
    def __init__(self):
        TestShiftedRotatedFunction.__init__(
            self,
            fbias=fbias_data[9],
            x_shift=rastrigin_func_data,
            rotate_M_D2=rastrigin_M_D2,
            rotate_M_D10=rastrigin_M_D10,
            rotate_M_D30=rastrigin_M_D30,
            rotate_M_D50=rastrigin_M_D50)


# CEC05 #11
class ShiftedRotatedWeierstrass(TestShiftedRotatedFunction,
                                Weierstrass):
    def __init__(self):
        TestShiftedRotatedFunction.__init__(
            self,
            fbias=fbias_data[10],
            x_shift=weierstrass_data,
            rotate_M_D2=weierstrass_M_D2,
            rotate_M_D10=weierstrass_M_D10,
            rotate_M_D30=weierstrass_M_D30,
            rotate_M_D50=weierstrass_M_D50)


# CEC05 #12
class Schwefel2_13(TestShiftedFunction):
    def __init__(self):
        TestShiftedFunction.__init__(self,
                                     fbias=fbias_data[11],
                                     x_shift=alpha_213)

    def __call__(self, x):
        return self.f(x) + self.fbias

    def f(self, x):
        a = a_213[:x.shape[1], :x.shape[1]]
        b = b_213[:x.shape[1], :x.shape[1]]
        alpha = self.x_shift[:x.shape[1]]

        A = np.sum(a*np.sin(alpha)[np.newaxis:,]
                   + b*np.cos(alpha)[np.newaxis:,], axis=1)

        B1 = a[np.newaxis]*np.sin(x[:, np.newaxis])
        B2 = b[np.newaxis]*np.cos(x[:, np.newaxis])
        B = np.sum(B1 + B2, axis=-1)

        return np.sum((A[np.newaxis:,] - B)**2, axis=-1)


# CEC05 #13
class ShiftedExpandedGriewankRosenbrock(TestShiftedFunction):
    def __init__(self):
        TestShiftedFunction.__init__(
            self,
            fbias=fbias_data[12],
            x_shift=EF8F2_func_data)
        self.rosenbrock_f = Rosenbrock()
        self.griewank_f = Griewank()

    def shift(self, x):
        shape = x.shape
        axis = [1]*(len(shape)-1) + [-1]
        return x - self.x_shift[:shape[-1]].reshape(axis) + 1

    def f(self, x):
        indexes = np.kron(
            np.arange(1, x.shape[1]-1, dtype=np.int64), np.array([1, 1]))
        indexes = np.insert(indexes, [0], [0])
        indexes = np.append(indexes, np.array([x.shape[1]-1, x.shape[1]-1, 0]))

        x_indexes = x[:, indexes]

        vertical_X = x_indexes.reshape(-1, 2)
        F2 = self.rosenbrock_f(vertical_X)
        F8 = self.griewank_f(F2.reshape(-1, 1))
        horizontal_X = F8.reshape(x.shape[0], -1)
        return np.sum(horizontal_X, axis=-1)


# CEC05 #14
class ShiftedRotatedExpandedScaffes_F6(TestShiftedRotatedFunction):
    def __init__(self):
        TestShiftedRotatedFunction.__init__(
            self,
            fbias=fbias_data[13],
            x_shift=E_ScafferF6_func_data,
            rotate_M_D2=E_ScafferF6_M_D2,
            rotate_M_D10=E_ScafferF6_M_D10,
            rotate_M_D30=E_ScafferF6_M_D30,
            rotate_M_D50=E_ScafferF6_M_D50)

    def Scaffes_F6(self, x):
        sum_of_power = x[:, 0]**2 + x[:, 1]**2
        up = np.sin(np.sqrt(sum_of_power))**2 - 0.5
        down = (1 + 0.001*(sum_of_power))**2
        return 0.5 + up/down

    def f(self, x):
        indexes = np.kron(
            np.arange(1, x.shape[1]-1, dtype=np.int64), np.array([1, 1]))
        indexes = np.insert(indexes, [0], [0])
        indexes = np.append(indexes, np.array([x.shape[1]-1, x.shape[1]-1, 0]))
        x_indexes = x[:, indexes]

        vertical_X = x_indexes.reshape(-1, 2)
        F = self.Scaffes_F6(vertical_X)
        horizontal_X = F.reshape(x.shape[0], -1)

        return np.sum(horizontal_X, axis=-1)


# CEC05 #15
class HybridCompositionFunction1(SampleHybridCompositionFunction):
    def __init__(self):
        SampleHybridCompositionFunction.__init__(
            self,
            basic_functions=(Rastrigin,
                             Rastrigin,
                             Weierstrass,
                             Weierstrass,
                             Griewank,
                             Griewank,
                             Ackley,
                             Ackley,
                             Sphere,
                             Sphere),
            sigmas=np.ones(10),
            lambdas=np.array([1, 1, 10, 10, 5/60, 5/60, 5/32,
                             5/32, 5/100, 5/100], dtype=np.float64),
            fbias=fbias_data[14],
            x_shift=hybrid_func1_data,
            M_D2=np.vstack([np.eye(2) for _ in range(10)]),
            M_D10=np.vstack([np.eye(10) for _ in range(10)]),
            M_D30=np.vstack([np.eye(30) for _ in range(10)]),
            M_D50=np.vstack([np.eye(50) for _ in range(10)]))


# CEC05 #16
class RotatedVersionHybridCompositionFunction1(SampleHybridCompositionFunction):
    def __init__(self):
        SampleHybridCompositionFunction.__init__(
            self,
            basic_functions=(Rastrigin,
                             Rastrigin,
                             Weierstrass,
                             Weierstrass,
                             Griewank,
                             Griewank,
                             Ackley,
                             Ackley,
                             Sphere,
                             Sphere),
            sigmas=np.ones(10),
            lambdas=np.array([1, 1, 10, 10, 5/60, 5/60, 5/32,
                             5/32, 5/100, 5/100], dtype=np.float64),
            fbias=fbias_data[15],
            x_shift=hybrid_func1_data,
            M_D2=hybrid_func1_M_D2,
            M_D10=hybrid_func1_M_D10,
            M_D30=hybrid_func1_M_D30,
            M_D50=hybrid_func1_M_D50)


# CEC05 #17
class RotatedVersionHybridCompositionFunction1Noise(SampleHybridCompositionFunction):
    def __init__(self):
        SampleHybridCompositionFunction.__init__(
            self,
            basic_functions=(Rastrigin,
                             Rastrigin,
                             Weierstrass,
                             Weierstrass,
                             Griewank,
                             Griewank,
                             Ackley,
                             Ackley,
                             Sphere,
                             Sphere),
            sigmas=np.ones(10),
            lambdas=np.array([1, 1, 10, 10, 5/60, 5/60, 5/32,
                             5/32, 5/100, 5/100], dtype=np.float64),
            fbias=fbias_data[16],
            x_shift=hybrid_func1_data,
            M_D2=hybrid_func1_M_D2,
            M_D10=hybrid_func1_M_D10,
            M_D30=hybrid_func1_M_D30,
            M_D50=hybrid_func1_M_D50)

    def __call__(self, x):
        G_x = RotatedVersionHybridCompositionFunction1()(x) - fbias_data[15]
        noise = np.abs(np.random.normal(size=G_x.shape))
        return G_x*(1 + 0.2*noise) + fbias_data[16]


# CEC05 #18
class RotatedHybridCompositionFunction(SampleHybridCompositionFunction):
    def __init__(self):
        SampleHybridCompositionFunction.__init__(
            self,
            basic_functions=(Ackley,
                             Ackley,
                             Rastrigin,
                             Rastrigin,
                             Sphere,
                             Sphere,
                             Weierstrass,
                             Weierstrass,
                             Griewank,
                             Griewank),
            sigmas=np.array([1, 2, 1.5, 1.5, 1, 1, 1.5,
                            1.5, 2, 2], dtype=np.float64),
            lambdas=np.array([2*(5/32), 5/32, 2*1, 1, 2*(5/100),
                             5/100, 2*10, 10, 2*(5/60), 5/60], dtype=np.float64),
            fbias=fbias_data[17],
            x_shift=hybrid_func2_data,
            M_D2=hybrid_func2_M_D2,
            M_D10=hybrid_func2_M_D10,
            M_D30=hybrid_func2_M_D30,
            M_D50=hybrid_func2_M_D50)
        self.x_shift[9] = 0


# CEC05 #19
class RotatedHybridCompositionFunctionNarrowBasin(SampleHybridCompositionFunction):
    def __init__(self):
        SampleHybridCompositionFunction.__init__(self, basic_functions=(Ackley,
                                                                        Ackley,
                                                                        Rastrigin,
                                                                        Rastrigin,
                                                                        Sphere,
                                                                        Sphere,
                                                                        Weierstrass,
                                                                        Weierstrass,
                                                                        Griewank,
                                                                        Griewank),
                                                 sigmas=np.array(
                                                     [0.1, 2, 1.5, 1.5, 1, 1, 1.5, 1.5, 2, 2], dtype=np.float64),
                                                 lambdas=np.array(
                                                     [0.1*(5/32), 5/32, 2*1, 1, 2*(5/100), 5/100, 2*10, 10, 2*(5/60), 5/60], dtype=np.float64),
                                                 fbias=fbias_data[17],
                                                 x_shift=hybrid_func2_data,
                                                 M_D2=hybrid_func2_M_D2,
                                                 M_D10=hybrid_func2_M_D10,
                                                 M_D30=hybrid_func2_M_D30,
                                                 M_D50=hybrid_func2_M_D50)
        self.x_shift[9] = 0


# CEC05 #20
class RotatedHybridCompositionFunctionOptimalBounds(SampleHybridCompositionFunction):
    def __init__(self):
        SampleHybridCompositionFunction.__init__(
            self, basic_functions=(Ackley,
                                   Ackley,
                                   Rastrigin,
                                   Rastrigin,
                                   Sphere,
                                   Sphere,
                                   Weierstrass,
                                   Weierstrass,
                                   Griewank,
                                   Griewank),
            sigmas=np.array([1, 2, 1.5, 1.5, 1, 1, 1.5,
                            1.5, 2, 2], dtype=np.float64),
            lambdas=np.array([2*(5/32), 5/32, 2*1, 1, 2*(5/100),
                             5/100, 2*10, 10, 2*(5/60), 5/60], dtype=np.float64),
            fbias=fbias_data[19],
            x_shift=hybrid_func2_data,
            M_D2=hybrid_func2_M_D2,
            M_D10=hybrid_func2_M_D10,
            M_D30=hybrid_func2_M_D30,
            M_D50=hybrid_func2_M_D50)
        self.x_shift[9] = 0

    def __call__(self, x):
        D = x.shape[1]
        self.x_shift[0][1:int(D/2):2] = 5
        return super().__call__(x)


# CEC05 #21
class HybridCompositionFunction3(SampleHybridCompositionFunction):
    def __init__(self):
        SampleHybridCompositionFunction.__init__(
            self,
            basic_functions=(ExpandedScaffers_F6,
                             ExpandedScaffers_F6,
                             Rastrigin,
                             Rastrigin,
                             F8F2,
                             F8F2,
                             Weierstrass,
                             Weierstrass,
                             Griewank,
                             Griewank),
            sigmas=np.array([1., 1., 1., 1., 1., 2., 2.,
                            2., 2., 2.], dtype=np.float64),
            lambdas=np.array([5*5/100, 5/100, 5*1, 1, 5*1, 1,
                             5*10, 10, 5*5/200, 5/200], dtype=np.float64),
            fbias=fbias_data[20],
            x_shift=hybrid_func3_data,
            M_D2=hybrid_func3_M_D2,
            M_D10=hybrid_func3_M_D10,
            M_D30=hybrid_func3_M_D30,
            M_D50=hybrid_func3_M_D50)


# CEC05 #22
class HybridCompositionFunction3H(SampleHybridCompositionFunction):
    def __init__(self):
        SampleHybridCompositionFunction.__init__(
            self,
            basic_functions=(ExpandedScaffers_F6,
                             ExpandedScaffers_F6,
                             Rastrigin,
                             Rastrigin,
                             F8F2,
                             F8F2,
                             Weierstrass,
                             Weierstrass,
                             Griewank,
                             Griewank),
            sigmas=np.array([1., 1., 1., 1., 1., 2., 2.,
                            2., 2., 2.], dtype=np.float64),
            lambdas=np.array([5*5/100, 5/100, 5*1, 1, 5*1, 1,
                             5*10, 10, 5*5/200, 5/200], dtype=np.float64),
            fbias=fbias_data[21],
            x_shift=hybrid_func3_data,
            M_D2=hybrid_func3_HM_D2,
            M_D10=hybrid_func3_HM_D10,
            M_D30=hybrid_func3_HM_D30,
            M_D50=hybrid_func3_HM_D50)


# CEC05 #23
class NonContinuousHybridCompositionFunction3(SampleHybridCompositionFunction):
    def __init__(self):
        SampleHybridCompositionFunction.__init__(
            self,
            basic_functions=(ExpandedScaffers_F6,
                             ExpandedScaffers_F6,
                             Rastrigin,
                             Rastrigin,
                             F8F2,
                             F8F2,
                             Weierstrass,
                             Weierstrass,
                             Griewank,
                             Griewank),
            sigmas=np.array([1., 1., 1., 1., 1., 2., 2.,
                            2., 2., 2.], dtype=np.float64),
            lambdas=np.array([5*5/100, 5/100, 5*1, 1, 5*1, 1,
                             5*10, 10, 5*5/200, 5/200], dtype=np.float64),
            fbias=fbias_data[22],
            x_shift=hybrid_func3_data,
            M_D2=hybrid_func3_M_D2,
            M_D10=hybrid_func3_M_D10,
            M_D30=hybrid_func3_M_D30,
            M_D50=hybrid_func3_M_D50)

    def __call__(self, x):
        shape = x.shape
        axis = [1]*(len(shape)-1) + [-1]
        o = self.x_shift[0][:shape[-1]].reshape(axis)
        cond = np.abs(x - o) >= 0.5
        x[cond] = (np.round(2*x)/2)[cond]
        return super().__call__(x)


# CEC05 #24
class HybridCompositionFunction4(SampleHybridCompositionFunction):
    def __init__(self):
        SampleHybridCompositionFunction.__init__(
            self,
            basic_functions=(Weierstrass,
                             ExpandedScaffers_F6,
                             F8F2,
                             Ackley,
                             Rastrigin,
                             Griewank,
                             NonContinuosExpandedScaffers_F6,
                             NonContinuosRastrigin,
                             HighConditionedElliptic,
                             SphereWithNoise),
            sigmas=np.array([2., 2., 2., 2., 2., 2., 2.,
                            2., 2., 2.], dtype=np.float64),
            lambdas=np.array([10, 5/20, 1, 5/32, 1, 5/100,
                             5/50, 1, 5/100, 5/100], dtype=np.float64),
            fbias=fbias_data[23],
            x_shift=hybrid_func4_data,
            M_D2=hybrid_func4_M_D2,
            M_D10=hybrid_func4_M_D10,
            M_D30=hybrid_func4_M_D30,
            M_D50=hybrid_func4_M_D50)


# CEC05 #25
class HybridCompositionFunction4withoutbounds(SampleHybridCompositionFunction):
    def __init__(self):
        SampleHybridCompositionFunction.__init__(
            self,
            basic_functions=(Weierstrass,
                             ExpandedScaffers_F6,
                             F8F2,
                             Ackley,
                             Rastrigin,
                             Griewank,
                             NonContinuosExpandedScaffers_F6,
                             NonContinuosRastrigin,
                             HighConditionedElliptic,
                             SphereWithNoise),
            sigmas=np.array([2., 2., 2., 2., 2., 2., 2.,
                            2., 2., 2.], dtype=np.float64),
            lambdas=np.array([10, 5/20, 1, 5/32, 1, 5/100,
                             5/50, 1, 5/100, 5/100], dtype=np.float64),
            fbias=fbias_data[24],
            x_shift=hybrid_func4_data,
            M_D2=hybrid_func4_M_D2,
            M_D10=hybrid_func4_M_D10,
            M_D30=hybrid_func4_M_D30,
            M_D50=hybrid_func4_M_D50)


import numpy as np
from ._base import TheFittest
from ._base import LastBest
from functools import partial
from ._differentialevolution import DifferentialEvolution
from ..tools.operators import binomial
from ..tools.generators import cauchy_distribution
from ..tools.transformations import lehmer_mean


class StatisticsSHADE:
    def __init__(self, mode='quick'):
        self.mode = mode
        self.population_g = np.array([])
        self.fitness = np.array([])
        self.H_F = np.array([], dtype=float)
        self.H_CR = np.array([], dtype=float)

    def append_arr(self, arr_to, arr_from):
        shape_to = (-1, arr_from.shape[0], arr_from.shape[1])
        shape_from = (1, arr_from.shape[0], arr_from.shape[1])
        result = np.vstack([arr_to.reshape(shape_to),
                            arr_from.copy().reshape(shape_from)])
        return result

    def update(self,
               population_g_i,
               fitness_i,
               H_F_i,
               H_CR_i):
        if self.mode == 'quick':
            self.fitness = np.append(self.fitness, np.max(fitness_i))

        elif self.mode == 'full':
            self.fitness = np.append(self.fitness, np.max(fitness_i))
            self.population_g = self.append_arr(self.population_g,
                                                population_g_i)
            self.H_F = H_F_i.copy().reshape(1, -1)
            self.H_CR = H_CR_i.copy().reshape(1, -1)
        else:
            raise ValueError('the "mode" must be either "quick" or "full"')
        return self


class SHADE(DifferentialEvolution):
    '''Tanabe, Ryoji & Fukunaga, Alex. (2013). Success-history based parameter adaptation
    for Differential Evolution. 2013 IEEE Congress on Evolutionary Computation,
    CEC 2013. 71-78. 10.1109/CEC.2013.6557555. '''

    def __init__(self,
                 fitness_function,
                 genotype_to_phenotype,
                 iters,
                 pop_size,
                 left,
                 right,
                 optimal_value = None,
                 termination_error_value = 0.,
                 no_increase_num = None,
                 minimization = False,
                 show_progress_each = None,
                 keep_history = None):
        DifferentialEvolution.__init__(
            self,
            fitness_function=fitness_function,
            genotype_to_phenotype=genotype_to_phenotype,
            iters=iters,
            pop_size=pop_size,
            left=left,
            right=right,
            optimal_value=optimal_value,
            termination_error_value=termination_error_value,
            no_increase_num=no_increase_num,
            minimization=minimization,
            show_progress_each=show_progress_each,
            keep_history=keep_history)

        self.thefittest: TheFittest
        self.stats: StatisticsSHADE
        self.H_size = pop_size

    def append_archive(self, archive, worse_i):
        archive = np.append(archive, worse_i, axis=0)
        if len(archive) > self.pop_size:
            np.random.shuffle(archive)
            archive = archive[:self.pop_size]
        return archive

    def current_to_pbest_1_archive(self, current, population, F_value, pop_archive):
        p_min = 2/len(population)
        p_i = np.random.uniform(p_min, 0.2)
        value = int(p_i*len(population))
        pbest = population[-value:]
        p_best_ind = np.random.randint(0, len(pbest))
        best = pbest[p_best_ind]

        r1 = np.random.choice(range(len(population)), size=1, replace=False)[0]
        r2 = np.random.choice(range(len(pop_archive)),
                              size=1, replace=False)[0]
        return current + F_value*(best - current) + F_value*(population[r1] - pop_archive[r2])

    def set_strategy(self,
                     elitism_param = True,
                     initial_population = None):
        self.update_pool()
        self.elitism = elitism_param
        self.initial_population = initial_population
        return self

    def mutation_and_crossover(self, popuation_g, popuation_g_archive, individ_g, F_i, CR_i):
        mutant = self.current_to_pbest_1_archive(individ_g, popuation_g, F_i,
                                                 popuation_g_archive)

        mutant_cr_g = binomial(individ_g, mutant, CR_i)
        mutant_cr_g = self.bounds_control(mutant_cr_g)
        return mutant_cr_g

    def evaluate_and_selection(self, mutant_cr_g, population_g, population_ph, fitness):
        offspring_g = population_g.copy()
        offspring_ph = population_ph.copy()
        offspring_fit = fitness.copy()

        mutant_cr_ph = self.genotype_to_phenotype(mutant_cr_g)
        mutant_cr_fit = self.evaluate(mutant_cr_ph)
        mask_more_equal = mutant_cr_fit >= fitness
        offspring_g[mask_more_equal] = mutant_cr_g[mask_more_equal]
        offspring_ph[mask_more_equal] = mutant_cr_ph[mask_more_equal]
        offspring_fit[mask_more_equal] = mutant_cr_fit[mask_more_equal]
        mask_more = mutant_cr_fit > fitness
        return offspring_g, offspring_ph, offspring_fit, mask_more

    def bounds_control(self, individ_g):
        low_mask = individ_g < self.left
        high_mask = individ_g > self.right

        individ_g[low_mask] = (self.left[low_mask] + individ_g[low_mask])/2
        individ_g[high_mask] = (self.right[high_mask] + individ_g[high_mask])/2
        return individ_g

    def generate_F_CR(self, H_F_i, H_CR_i, size):
        F_i = np.zeros(size)
        CR_i = np.zeros(size)
        for i in range(size):
            r_i = np.random.randint(0, len(H_F_i))
            u_F = H_F_i[r_i]
            u_CR = H_CR_i[r_i]
            F_i[i] = self.randc01(u_F)
            CR_i[i] = self.randn01(u_CR)
        return F_i, CR_i

    def randc01(self, u):
        value = cauchy_distribution(loc=u, scale=0.1)[0]
        while value <= 0:
            value = cauchy_distribution(loc=u, scale=0.1)[0]
        if value > 1:
            value = 1
        return value

    def randn01(self, u):
        value = np.random.normal(u, 0.1)
        if value < 0:
            value = 0
        elif value > 1:
            value = 1
        return value

    def update_u_F(self, u_F, S_F):
        if len(S_F):
            return lehmer_mean(S_F)
        return u_F

    def update_u_CR(self, u_CR, S_CR, df):
        if len(S_CR):
            sum_ = np.sum(df)
            if sum_ > 0:
                weight_i = df/sum_
                return np.sum(weight_i*S_CR)
        return u_CR

    def fit(self):
        H_F = np.full(self.H_size, 0.5)
        H_CR = np.full(self.H_size, 0.5)
        k = 0
        next_k = 1

        external_archive = np.zeros(shape=(0, len(self.left)))

        population_g = self.generate_init_pop()
        population_ph = self.genotype_to_phenotype(population_g)
        fitness = self.evaluate(population_ph)

        argsort = np.argsort(fitness)
        population_g = population_g[argsort]
        population_ph = population_ph[argsort]
        fitness = fitness[argsort]

        self.thefittest = TheFittest().update(population_g,
                                              population_ph,
                                              fitness)
        lastbest = LastBest().update(self.thefittest.fitness)
        if self.keep_history is not None:
            self.stats = StatisticsSHADE(
                mode=self.keep_history).update(population_g,
                                               fitness,
                                               H_F,
                                               H_CR)
        for i in range(self.iters-1):
            self.show_progress(i)
            if self.termitation_check(lastbest.no_increase):
                break
            else:
                F_i, CR_i = self.generate_F_CR(H_F, H_CR, self.pop_size)
                pop_archive = np.vstack(
                    [population_g.copy(), external_archive.copy()])

                partial_mut_and_cross = partial(self.mutation_and_crossover,
                                                population_g.copy(), pop_archive.copy())
                mutant_cr_g = np.array(list(map(partial_mut_and_cross,
                                                population_g.copy(),
                                                F_i.copy(), CR_i.copy())))

                stack = self.evaluate_and_selection(mutant_cr_g.copy(),
                                                    population_g.copy(),
                                                    population_ph.copy(),
                                                    fitness.copy())

                succeses = stack[3]
                will_be_replaced_pop = population_g[succeses].copy()
                will_be_replaced_fit = fitness[succeses].copy()
                s_F = F_i[succeses].copy()
                s_CR = CR_i[succeses].copy()

                external_archive = self.append_archive(
                    external_archive, will_be_replaced_pop)

                population_g = stack[0]
                population_ph = stack[1]
                fitness = stack[2]

                df = np.abs(will_be_replaced_fit - fitness[succeses])

                if self.elitism:
                    population_g[-1], population_ph[-1], fitness[-1] = self.thefittest.get()
                argsort = np.argsort(fitness)
                population_g = population_g[argsort]
                population_ph = population_ph[argsort]
                fitness = fitness[argsort]

                if next_k == self.H_size:
                    next_k = 0

                H_F[next_k] = self.update_u_F(H_F[k], s_F)
                H_CR[next_k] = self.update_u_CR(H_CR[k], s_CR, df)

                self.thefittest.update(population_g, population_ph, fitness)
                lastbest.update(self.thefittest.fitness)

                if self.keep_history is not None:
                    self.stats.update(population_g,
                                      fitness,
                                      H_F,
                                      H_CR)

                if k == self.H_size - 1:
                    k = 0
                    next_k = 1
                else:
                    k += 1
                    next_k += 1

        return self



import numpy as np
from functools import partial
from ._base import TheFittest
from ._base import LastBest
from ._base import EvolutionaryAlgorithm
from ..tools.operators import proportional_selection
from ..tools.operators import rank_selection
from ..tools.operators import tournament_selection
from ..tools.operators import empty_crossover
from ..tools.operators import one_point_crossover
from ..tools.operators import two_point_crossover
from ..tools.operators import uniform_crossover
from ..tools.operators import uniform_prop_crossover
from ..tools.operators import uniform_rank_crossover
from ..tools.operators import uniform_tour_crossover
from ..tools.operators import flip_mutation
from ..tools.generators import binary_string_population
from ..tools.transformations import scale_data
from ..tools.transformations import rank_data


class Statistics:
    def __init__(self, mode='quick'):
        self.mode = mode
        self.population_g = np.array([])
        self.fitness = np.array([])

    def append_arr(self, arr_to, arr_from):
        shape_to = (-1, arr_from.shape[0], arr_from.shape[1])
        shape_from = (1, arr_from.shape[0], arr_from.shape[1])
        result = np.vstack([arr_to.reshape(shape_to),
                            arr_from.copy().reshape(shape_from)])
        return result

    def update(self,
               population_g_i,
               fitness_i):
        if self.mode == 'quick':
            self.fitness = np.append(self.fitness, np.max(fitness_i))
        elif self.mode == 'full':
            self.fitness = np.append(self.fitness, np.max(fitness_i))
            self.population_g = self.append_arr(self.population_g,
                                                population_g_i)
        else:
            raise ValueError('the "mode" must be either "quick" or "full"')
        return self


class GeneticAlgorithm(EvolutionaryAlgorithm):
    def __init__(self,
                 fitness_function,
                 genotype_to_phenotype,
                 iters,
                 pop_size,
                 str_len,
                 optimal_value = None,
                 termination_error_value = 0.,
                 no_increase_num = None,
                 minimization = False,
                 show_progress_each = None,
                 keep_history = None):
        EvolutionaryAlgorithm.__init__(
            self,
            fitness_function=fitness_function,
            genotype_to_phenotype=genotype_to_phenotype,
            iters=iters,
            pop_size=pop_size,
            optimal_value=optimal_value,
            termination_error_value=termination_error_value,
            no_increase_num=no_increase_num,
            minimization=minimization,
            show_progress_each=show_progress_each,
            keep_history=keep_history)

        self.str_len = str_len
        self.thefittest: TheFittest
        self.stats: Statistics
        self.s_pool: dict
        self.c_pool: dict
        self.m_pool: dict
        self.initial_population: np.ndarray
        self.tour_size: int
        self.elitism: bool
        self.parents_num: int
        self.mutation_rate: float
        self.s_set: tuple
        self.c_set: tuple
        self.m_set: tuple

        self.set_strategy()

    def generate_init_pop(self):
        if self.initial_population is None:
            population_g = binary_string_population(
                self.pop_size, self.str_len)
        else:
            population_g = self.initial_population
        return population_g

    def update_pool(self):
        self.s_pool = {'proportional': (proportional_selection, None),
                       'rank': (rank_selection, None),
                       'tournament_k': (tournament_selection, self.tour_size),
                       'tournament_3': (tournament_selection, 3),
                       'tournament_5': (tournament_selection, 5),
                       'tournament_7': (tournament_selection, 7)}

        self.c_pool = {'empty': (empty_crossover, 1),
                       'one_point': (one_point_crossover, 2),
                       'two_point': (two_point_crossover, 2),
                       'uniform2': (uniform_crossover, 2),
                       'uniform7': (uniform_crossover, 7),
                       'uniformk': (uniform_crossover, self.parents_num),
                       'uniform_prop2': (uniform_prop_crossover, 2),
                       'uniform_prop7': (uniform_prop_crossover, 7),
                       'uniform_propk': (uniform_prop_crossover, self.parents_num),
                       'uniform_rank2': (uniform_rank_crossover, 2),
                       'uniform_rank7': (uniform_rank_crossover, 7),
                       'uniform_rankk': (uniform_rank_crossover, self.parents_num),
                       'uniform_tour3': (uniform_tour_crossover, 3),
                       'uniform_tour7': (uniform_tour_crossover, 7),
                       'uniform_tourk': (uniform_tour_crossover, self.parents_num)}

        self.m_pool = {'weak':  (flip_mutation, 1/(3*self.str_len)),
                       'average':  (flip_mutation, 1/(self.str_len)),
                       'strong': (flip_mutation, min(1, 3/self.str_len)),
                       'custom_rate': (flip_mutation, self.mutation_rate)}

    def set_strategy(self,
                     selection_oper = 'tournament_k',
                     crossover_oper = 'uniform2',
                     mutation_oper = 'weak',
                     tour_size_param = 2,
                     initial_population = None,
                     elitism_param = True,
                     parents_num_param = 7,
                     mutation_rate_param = 0.05):

        self.tour_size = tour_size_param
        self.initial_population = initial_population
        self.elitism = elitism_param
        self.parents_num = parents_num_param
        self.mutation_rate = mutation_rate_param

        self.update_pool()

        self.s_set = self.s_pool[selection_oper]
        self.c_set = self.c_pool[crossover_oper]
        self.m_set = self.m_pool[mutation_oper]

        return self

    def create_offspring(self, population_g, fitness_scale, fitness_rank, _):
        crossover_func, quantity = self.c_set
        selection_func, tour_size = self.s_set
        mutation_func, proba = self.m_set

        indexes = selection_func(fitness_scale,
                                 fitness_rank,
                                 tour_size,
                                 quantity)

        parents = population_g[indexes].copy()
        fitness_scale_p = fitness_scale[indexes].copy()
        fitness_rank_p = fitness_rank[indexes].copy()

        offspring_no_mutated = crossover_func(parents,
                                              fitness_scale_p,
                                              fitness_rank_p)

        mutant = mutation_func(offspring_no_mutated, proba)
        return mutant

    def fit(self):
        population_g = self.generate_init_pop()
        population_ph = self.genotype_to_phenotype(population_g)
        fitness = self.evaluate(population_ph)
        fitness_scale = scale_data(fitness)
        fitness_rank = rank_data(fitness)

        self.thefittest = TheFittest().update(population_g,
                                              population_ph,
                                              fitness)
        lastbest = LastBest().update(self.thefittest.fitness)
        if self.keep_history is not None:
            self.stats = Statistics(
                mode=self.keep_history).update(population_g,
                                               fitness)

        for i in range(self.iters-1):
            self.show_progress(i)
            if self.termitation_check(lastbest.no_increase):
                break
            else:
                partial_create_offspring = partial(self.create_offspring,
                                                   population_g,
                                                   fitness_scale,
                                                   fitness_rank)
                map_ = map(partial_create_offspring, range(self.pop_size))
                population_g = np.array(list(map_), dtype=np.byte)

                population_ph = self.genotype_to_phenotype(population_g)
                fitness = self.evaluate(population_ph)

                if self.elitism:
                    population_g[-1], population_ph[-1], fitness[-1] = self.thefittest.get()
                fitness_scale = scale_data(fitness)
                fitness_rank = rank_data(fitness)

                self.thefittest.update(population_g, population_ph, fitness)
                lastbest.update(self.thefittest.fitness)
                if self.keep_history is not None:
                    self.stats.update(population_g,
                                      fitness)
        return self



import numpy as np
from ._base import TheFittest
from ._base import LastBest
from functools import partial
from ._differentialevolution import DifferentialEvolution
from ..tools.operators import binomial
from ..tools.generators import cauchy_distribution
from ..tools.transformations import lehmer_mean


class StatisticsJADE:
    def __init__(self, mode='quick'):
        self.mode = mode
        self.population_g = np.array([])
        self.fitness = np.array([])
        self.u_F = np.array([], dtype=float)
        self.u_CR = np.array([], dtype=float)

    def append_arr(self, arr_to, arr_from):
        shape_to = (-1, arr_from.shape[0], arr_from.shape[1])
        shape_from = (1, arr_from.shape[0], arr_from.shape[1])
        result = np.vstack([arr_to.reshape(shape_to),
                            arr_from.copy().reshape(shape_from)])
        return result

    def update(self,
               population_g_i,
               fitness_i,
               u_F_i,
               U_CR_i):
        if self.mode == 'quick':
            self.fitness = np.append(self.fitness, np.max(fitness_i))
        elif self.mode == 'full':
            self.fitness = np.append(self.fitness, np.max(fitness_i))
            self.population_g = self.append_arr(self.population_g,
                                                population_g_i)
            self.u_F = np.append(self.u_F, u_F_i)
            self.u_CR = np.append(self.u_CR, U_CR_i)
        else:
            raise ValueError('the "mode" must be either "quick" or "full"')
        return self


class JADE(DifferentialEvolution):
    '''Zhang, Jingqiao & Sanderson, A.C.. (2009). JADE: Adaptive Differential Evolution With Optional External Archive.
     Evolutionary Computation, IEEE Transactions on. 13. 945 - 958. 10.1109/TEVC.2009.2014613. '''

    def __init__(self,
                 fitness_function,
                 genotype_to_phenotype,
                 iters,
                 pop_size,
                 left,
                 right,
                 optimal_value=None,
                 termination_error_value=0.,
                 no_increase_num=None,
                 minimization=False,
                 show_progress_each=None,
                 keep_history=None):
        DifferentialEvolution.__init__(
            self,
            fitness_function=fitness_function,
            genotype_to_phenotype=genotype_to_phenotype,
            iters=iters,
            pop_size=pop_size,
            left=left,
            right=right,
            optimal_value=optimal_value,
            termination_error_value=termination_error_value,
            no_increase_num=no_increase_num,
            minimization=minimization,
            show_progress_each=show_progress_each,
            keep_history=keep_history)

        self.thefittest: TheFittest
        self.stats: StatisticsJADE

        self.c = 0.1
        self.p = 0.05

    def set_strategy(self, c_param = 0.1,
                     p_param = 0.05,
                     elitism_param = True,
                     initial_population = None):
        self.update_pool()
        self.c = c_param
        self.p = p_param
        self.elitism = elitism_param
        self.initial_population = initial_population
        return self

    def current_to_pbest_1_archive(self, current, population, F_value, pop_archive):
        p_i = self.p
        value = int(p_i*len(population))
        pbest = population[-value:]
        p_best_ind = np.random.randint(0, len(pbest))
        best = pbest[p_best_ind]

        r1 = np.random.choice(range(len(population)), size=1, replace=False)[0]
        r2 = np.random.choice(range(len(pop_archive)),
                              size=1, replace=False)[0]
        return current + F_value*(best - current) + F_value*(population[r1] - pop_archive[r2])

    def mutation_and_crossover(self, popuation_g, popuation_g_archive, individ_g, F_i, CR_i):
        mutant = self.current_to_pbest_1_archive(individ_g, popuation_g, F_i,
                                                 popuation_g_archive)

        mutant_cr_g = binomial(individ_g, mutant, CR_i)
        mutant_cr_g = self.bounds_control(mutant_cr_g)
        return mutant_cr_g

    def evaluate_and_selection(self, mutant_cr_g, population_g, population_ph, fitness):
        offspring_g = population_g.copy()
        offspring_ph = population_ph.copy()
        offspring_fit = fitness.copy()

        mutant_cr_ph = self.genotype_to_phenotype(mutant_cr_g)
        mutant_cr_fit = self.evaluate(mutant_cr_ph)
        mask = mutant_cr_fit >= fitness
        offspring_g[mask] = mutant_cr_g[mask]
        offspring_ph[mask] = mutant_cr_ph[mask]
        offspring_fit[mask] = mutant_cr_fit[mask]
        return offspring_g, offspring_ph, offspring_fit, mask

    def bounds_control(self, individ_g):
        low_mask = individ_g < self.left
        high_mask = individ_g > self.right

        individ_g[low_mask] = (self.left[low_mask] + individ_g[low_mask])/2
        individ_g[high_mask] = (self.right[high_mask] + individ_g[high_mask])/2
        return individ_g

    def generate_F(self, u_F):
        F_i = cauchy_distribution(loc=u_F, scale=0.1, size=self.pop_size)
        mask = F_i <= 0
        while np.any(mask):
            F_i[mask] = cauchy_distribution(
                loc=u_F, scale=0.1, size=len(F_i[mask]))
            mask = F_i <= 0
        F_i[F_i >= 1] = 1
        return F_i

    def generate_CR(self, u_CR):
        CR_i = np.random.normal(u_CR, 0.1, self.pop_size)
        CR_i[CR_i >= 1] = 1
        CR_i[CR_i <= 0] = 0
        return CR_i

    def update_u_F(self, u_F, S_F):
        if len(S_F):
            u_F = (1 - self.c)*u_F + self.c*lehmer_mean(S_F)
        return u_F

    def update_u_CR(self, u_CR, S_CR):
        if len(S_CR):
            u_CR = (1 - self.c)*u_CR + self.c*np.mean(S_CR)
        return u_CR

    def append_archive(self, archive, worse_i):
        archive = np.append(archive, worse_i, axis=0)
        if len(archive) > self.pop_size:
            np.random.shuffle(archive)
            archive = archive[:self.pop_size]
        return archive

    def fit(self):
        u_F = u_CR = 0.5
        external_archive = np.zeros(shape=(0, len(self.left)))

        population_g = self.generate_init_pop()
        population_ph = self.genotype_to_phenotype(population_g)
        fitness = self.evaluate(population_ph)

        argsort = np.argsort(fitness)
        population_g = population_g[argsort]
        population_ph = population_ph[argsort]
        fitness = fitness[argsort]

        self.thefittest = TheFittest().update(population_g,
                                              population_ph,
                                              fitness)
        lastbest = LastBest().update(self.thefittest.fitness)
        if self.keep_history is not None:
            self.stats = StatisticsJADE(
                mode=self.keep_history).update(population_g,
                                               fitness,
                                               u_F,
                                               u_CR)

        for i in range(self.iters-1):
            self.show_progress(i)
            if self.termitation_check(lastbest.no_increase):
                break
            else:
                F_i = self.generate_F(u_F)
                CR_i = self.generate_CR(u_CR)
                pop_archive = np.vstack(
                    [population_g.copy(), external_archive.copy()])

                partial_mut_and_cross = partial(self.mutation_and_crossover,
                                                population_g.copy(), pop_archive.copy())
                mutant_cr_g = np.array(list(map(partial_mut_and_cross,
                                                population_g.copy(),
                                                F_i.copy(), CR_i.copy())))

                stack = self.evaluate_and_selection(mutant_cr_g.copy(),
                                                    population_g.copy(),
                                                    population_ph.copy(),
                                                    fitness.copy())

                succeses = stack[3]
                will_be_replaced = population_g[succeses].copy()
                s_F = F_i[succeses].copy()
                s_CR = CR_i[succeses].copy()

                external_archive = self.append_archive(
                    external_archive, will_be_replaced)

                population_g = stack[0]
                population_ph = stack[1]
                fitness = stack[2]

                if self.elitism:
                    population_g[-1], population_ph[-1], fitness[-1] = self.thefittest.get()
                argsort = np.argsort(fitness)
                population_g = population_g[argsort]
                population_ph = population_ph[argsort]
                fitness = fitness[argsort]

                u_F = self.update_u_F(u_F, s_F)
                u_CR = self.update_u_CR(u_CR, s_CR)

                self.thefittest.update(population_g, population_ph, fitness)
                lastbest.update(self.thefittest.fitness)

                if self.keep_history is not None:
                    self.stats.update(population_g,
                                      fitness,
                                      u_F,
                                      u_CR)

        return self
